# THIS FILE IS GENERATED FROM macrosynergy SETUP.PY
short_version: str = '0.0.32'
version: str = '0.0.32'
full_version: str = '0.0.32'
git_revision: str = '16ec103208b508468ce8a9ddc1feb2b54145b321'
release: bool = True
if not release:
    version = full_version



import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.ticker as mticker

from typing import List, Union, Tuple, Optional
from itertools import product
from macrosynergy.management.simulate_quantamental_data import make_qdf
from macrosynergy.management.shape_dfs import reduce_df
from macrosynergy.panel.make_zn_scores import make_zn_scores
from macrosynergy.management.update_df import update_df

class NaivePnL:

    """
    Computes and collects illustrative PnLs with limited signal options and
    disregarding transaction costs.

    :param <pd.Dataframe> df: standardized DataFrame with the following necessary
        columns: 'cid', 'xcat', 'real_date' and 'value'.
    :param <str> ret: return category.
    :param <List[str]> sigs: signal categories. Able to pass in multiple possible signals
        to the Class' constructor and their respective vintages will be held on the
        instance's DataFrame. The signals can subsequently be referenced through the
        self.make_pnl() method which receives a single signal per call.
    :param <List[str]> cids: cross sections that are traded. Default is all in the
        dataframe.
    :param <str, List[str]> bms: list of benchmark tickers for which
        correlations are displayed against PnL strategies.
    :param <str> start: earliest date in ISO format. Default is None and earliest date
        in df is used.
    :param <str> end: latest date in ISO format. Default is None and latest date in df
        is used.
    :param <dict> blacklist: cross-sections with date ranges that should be excluded
        from the dataframe.

    """
    def __init__(self, df: pd.DataFrame, ret: str, sigs: List[str],
                 cids: List[str] = None, bms: Union[str, List[str]] = None,
                 start: str = None, end: str = None, blacklist: dict = None):

        df["real_date"] = pd.to_datetime(df["real_date"], format="%Y-%m-%d")

        # Will host the benchmarks.
        dfd = df.copy()

        self.dfd = df
        assert isinstance(ret, str), "The return category expects a single <str>."
        self.ret = ret
        xcats = [ret] + sigs

        cols = ["cid", "xcat", "real_date", "value"]
        # Potentially excludes the benchmarks but will be held on the instance level
        # through self.dfd.
        df, self.xcats, self.cids = reduce_df(
            df[cols], xcats, cids, start, end, blacklist, out_all=True
        )

        self.df = df
        self.sigs = sigs

        ticker_func = lambda t: t[0] + "_" + t[1]
        self.tickers = list(map(ticker_func, product(self.cids, self.xcats)))

        self.df["real_date"] = pd.to_datetime(self.df["real_date"])

        # Data structure used to track all of the generated PnLs from make_pnl() method.
        self.pnl_names = []
        self.signal_df = {}
        self.black = blacklist

        self.bm_bool = isinstance(bms, (str, list))
        if self.bm_bool:
            bms = [bms] if isinstance(bms, str) else bms

            # Pass in the original DataFrame; negative signal will not have been applied
            # which will corrupt the use of the benchmark categories.
            bm_dict = self.add_bm(df=dfd, bms=bms, tickers=self.tickers)

            self._bm_dict = bm_dict

    def add_bm(self, df: pd.DataFrame, bms: List[str],
               tickers: List[str]):
        """
        Returns a dictionary with benchmark return series.

        :param <pd.DataFrame> df: aggregate DataFrame passed into the Class.
        :param <List[str]> bms: benchmark return tickers.
        :param <List[str]> tickers: the available tickers held in the reduced DataFrame.
            The reduced DataFrame consists exclusively of the signal & return categories.
        """

        bm_dict = {}

        for bm in bms:
            # Accounts for appending "_NEG" to the ticker.
            bm_s = bm.split("_", maxsplit=1)
            cid = bm_s[0]
            xcat = bm_s[1]
            dfa = df[(df['cid'] == cid) & (df['xcat'] == xcat)]

            if dfa.shape[0] == 0:
                print(f"{bm} has no observations in the DataFrame.")
            else:
                df_single_bm = dfa.pivot(
                    index='real_date', columns='xcat', values='value'
                )
                df_single_bm.columns = [bm]
                bm_dict[bm] = df_single_bm
                if bm not in tickers:
                    self.df = update_df(self.df, dfa)

        return bm_dict

    @classmethod
    def __make_signal__(cls, dfx: pd.DataFrame, sig: str, sig_op: str = 'zn_score_pan',
                        min_obs: int = 252, iis: bool = True, sequential: bool = True,
                        neutral: str = 'zero', thresh: float = None):
        """
        Helper function used to produce the raw signal that forms the basis for
        positioning.

        :param <pd.DataFrame> dfx: DataFrame defined over the return & signal category.
        :param <str> sig: name of the raw signal.
        :param <str> sig_op: signal transformation.
        :param <int> min_obs: the minimum number of observations required to calculate
            zn_scores. Default is 252.
        :param <bool> iis: if True (default) zn-scores are also calculated for the initial
            sample period defined by min_obs, on an in-sample basis, to avoid losing
            history.
        :param <bool> sequential: if True (default) score parameters are estimated
            sequentially with concurrently available information only.
        :param <str> neutral: method to determine neutral level.
        :param <float> thresh: threshold value beyond which scores are winsorized,

        """

        if sig_op == 'binary':
            dfw = dfx.pivot(index=['cid', 'real_date'], columns='xcat', values='value')
            dfw['psig'] = np.sign(dfw[sig])
        else:
            panw = 1 if sig_op == 'zn_score_pan' else 0
            # The re-estimation frequency for the neutral level and standard deviation
            # will be the same as the re-balancing frequency. For instance, if the
            # neutral level is computed weekly, a material change in the signal will only
            # manifest along a similar timeline. Therefore, re-estimation and
            # re-balancing frequencies match.
            df_ms = make_zn_scores(dfx, xcat=sig, neutral=neutral, pan_weight=panw,
                                   sequential=sequential, min_obs=min_obs, iis=iis,
                                   thresh=thresh)

            df_ms = df_ms.drop('xcat', axis=1)
            df_ms['xcat'] = 'psig'

            dfx_concat = pd.concat([dfx, df_ms])
            dfw = dfx_concat.pivot(index=['cid', 'real_date'], columns='xcat',
                                   values='value')

        # Reconstruct the DataFrame to recognise the signal's start date for each
        # individual cross-section
        dfw_list = []
        for c, cid_df in dfw.groupby(level=0):
            first_date = cid_df.loc[:, 'psig'].first_valid_index()
            cid_df = cid_df.loc[first_date:, :]
            dfw_list.append(cid_df)

        return pd.concat(dfw_list)

    @classmethod
    def rebalancing(cls, dfw: pd.DataFrame, rebal_freq: str = 'daily', rebal_slip = 0):
        """
        The signals are calculated daily and for each individual cross-section defined in
        the panel. However, re-balancing a position can occur more infrequently than
        daily. Therefore, produce the re-balancing values according to the more
        infrequent timeline (weekly or monthly).

        :param <pd.Dataframe> dfw: DataFrame with each category represented by a column
            and the daily signal is also included with the column name 'psig'.
        :param <str> rebal_freq: re-balancing frequency for positions according to signal
            must be one of 'daily' (default), 'weekly' or 'monthly'.
        :param <str> rebal_slip: re-balancing slippage in days.

        :return <pd.Series>: will return a pd.Series containing the associated signals
            according to the re-balancing frequency.
        """

        # The re-balancing days are the first of the respective time-periods because of
        # the shift forward by one day applied earlier in the code. Therefore, only
        # concerned with the minimum date of each re-balance period.
        dfw['year'] = dfw['real_date'].dt.year
        if rebal_freq == 'monthly':
            dfw['month'] = dfw['real_date'].dt.month
            rebal_dates = dfw.groupby(['cid', 'year', 'month'])['real_date'].min()
        elif rebal_freq == 'weekly':
            dfw['week'] = dfw['real_date'].dt.week
            rebal_dates = dfw.groupby(['cid', 'year', 'week'])['real_date'].min()

        # Convert the index, 'cid', to a formal column aligned to the re-balancing dates.
        r_dates_df = rebal_dates.reset_index(level=0)
        r_dates_df.reset_index(drop=True, inplace=True)
        dfw = dfw[['real_date', 'psig', 'cid']]

        # Isolate the required signals on the re-balancing dates. Only concerned with the
        # respective signal on the re-balancing date. However, the produced DataFrame
        # will only be defined over the re-balancing dates. Therefore, merge the
        # aforementioned DataFrame with the original DataFrame such that all business
        # days are included. The intermediary dates, dates between re-balancing dates,
        # will initially be populated by NA values. To ensure the signal is used for the
        # duration between re-balancing dates, forward fill the computed signal over the
        # associated dates.

        # The signal is computed for each individual cross-section. Therefore, merge on
        # the real_date and the cross-section.
        rebal_merge = r_dates_df.merge(dfw, how='left', on=['real_date', 'cid'])
        # Re-establish the daily date series index where the intermediary dates, between
        # the re-balancing dates, will be populated using a forward fill.
        rebal_merge = dfw[['real_date', 'cid']].merge(
            rebal_merge, how='left', on=['real_date', 'cid']
        )
        rebal_merge['psig'] = rebal_merge['psig'].fillna(method='ffill').shift(rebal_slip)
        rebal_merge = rebal_merge.sort_values(['cid', 'real_date'])

        rebal_merge = rebal_merge.set_index('real_date')
        sig_series = rebal_merge.drop(['cid'], axis=1)

        return sig_series

    def make_pnl(self, sig: str, sig_op: str = 'zn_score_pan', sig_neg: bool = False,
                 pnl_name: str = None,
                 rebal_freq: str = 'daily', rebal_slip = 0, vol_scale: float = None,
                 min_obs: int = 261, iis: bool = True, sequential: bool = True,
                 neutral: str = 'zero', thresh: float = None):

        """
        Calculate daily PnL and add to class instance.

        :param <str> sig: name of raw signal that is basis for positioning. The signal
            is assumed to be recorded at the end of the day prior to position taking.
        :param <str> sig_op: signal transformation options; must be one of
            'zn_score_pan', 'zn_score_cs', or 'binary'. The default is 'zn_score_pan'.
            'zn_score_pan': transforms raw signals into z-scores around zero value
            based on the whole panel. The neutral level & standard deviation will use the
            cross-section of panels.
            'zn_score_cs': transforms signals to z-scores around zero based on
            cross-section alone.
            'binary': transforms signals into uniform long/shorts (1/-1) across all
            sections.
            N.B.: zn-score here means standardized score with zero being the natural
            neutral level and standardization through division by mean absolute value.
        :param <str> sig_neg: if True the PnL is based on the negative value of the
            transformed signal. Default is False.
        :param <str> pnl_name: name of the PnL to be generated and stored.
            Default is None, i.e. a default name is given. The default name will be:
            'PNL_<signal name>[_<NEG>]', with the last part added if sig_neg has been
            set to True.
            Previously calculated PnLs of the same name will be overwritten. This means
            that if a set of PnLs are to be compared, each PnL requires a distinct name.
        :param <str> rebal_freq: re-balancing frequency for positions according to signal
            must be one of 'daily' (default), 'weekly' or 'monthly'. The re-balancing is
            only concerned with the signal value on the re-balancing date which is
            delimited by the frequency chosen.
            Additionally, the re-balancing frequency will be applied to make_zn_scores()
            if used as the method to produce the raw signals.
        :param <str> rebal_slip: re-balancing slippage in days. Default is 1 which
            means that it takes one day to re-balance the position and that the new
            positions produce PnL from the second day after the signal has been recorded.
        :param <bool> vol_scale: ex-post scaling of PnL to annualized volatility given.
            This is for comparative visualization and not out-of-sample. Default is none.
        :param <int> min_obs: the minimum number of observations required to calculate
            zn_scores. Default is 252.
        :param <bool> iis: if True (default) zn-scores are also calculated for the initial
            sample period defined by min_obs, on an in-sample basis, to avoid losing
            history.
        :param <bool> sequential: if True (default) score parameters (neutral level and
            standard deviations) are estimated sequentially with concurrently available
            information only.
        :param <str> neutral: method to determine neutral level. Default is 'zero'.
            Alternatives are 'mean' and "median".
        :param <float> thresh: threshold value beyond which scores are winsorized,
            i.e. contained at that threshold. Therefore, the threshold is the maximum
            absolute score value that the function is allowed to produce. The minimum
            threshold is one standard deviation. Default is no threshold.

        """

        # A. Checks

        error_sig = f"Signal category missing from the options defined on the class: " \
                    f"{self.sigs}. "
        assert sig in self.sigs, error_sig
        sig_options = ['zn_score_pan', 'zn_score_cs', 'binary']
        error_sig_method = f"The signal transformation method, {sig_op}, is not one of " \
                           f"the options specified: {sig_options}."
        assert sig_op in sig_options, error_sig_method
        freq_params = ['daily', 'weekly', 'monthly']
        freq_error = f"Re-balancing frequency must be one of: {freq_params}."
        assert rebal_freq in freq_params, freq_error

        sig_neg_error = "Boolean object expected for negative conversion."
        assert isinstance(sig_neg, bool), sig_neg_error

        # B. Extract DataFrame of exclusively return and signal categories in time series
        # format.
        dfx = self.df[self.df['xcat'].isin([self.ret, sig])]

        dfw = self.__make_signal__(
            dfx=dfx, sig=sig, sig_op=sig_op, min_obs=min_obs, iis=iis,
            sequential=sequential, neutral=neutral, thresh=thresh
        )

        if sig_neg:
            dfw['psig'] *= -1
            neg = "_NEG"
        else:
            neg = ""

        # Multi-index DataFrame with a natural minimum lag applied.
        dfw['psig'] = dfw['psig'].groupby(level=0).shift(1)
        dfw.reset_index(inplace=True)
        dfw = dfw.rename_axis(None, axis=1)

        dfw = dfw.sort_values(['cid', 'real_date'])

        if rebal_freq != 'daily':
            sig_series = self.rebalancing(dfw=dfw, rebal_freq=rebal_freq,
                                          rebal_slip=rebal_slip)
            dfw['sig'] = np.squeeze(sig_series.to_numpy())
        else:
            dfw = dfw.rename({'psig': 'sig'}, axis=1)

        # The signals are generated across the panel.
        dfw['value'] = dfw[self.ret] * dfw['sig']

        df_pnl = dfw.loc[:, ['cid', 'real_date', 'value']]

        # Compute the return across the panel. The returns are still computed daily
        # regardless of the re-balancing frequency potentially occurring weekly or
        # monthly.
        df_pnl_all = df_pnl.groupby(['real_date']).sum(numeric_only=True)
        df_pnl_all = df_pnl_all[df_pnl_all['value'].cumsum() != 0]
        # Returns are computed for each cross-section and across the panel.
        df_pnl_all['cid'] = 'ALL'
        df_pnl_all = df_pnl_all.reset_index()[df_pnl.columns]
        # Will be inclusive of each individual cross-section's signal-adjusted return and
        # the aggregated panel return.
        df_pnl = pd.concat([df_pnl, df_pnl_all])

        if vol_scale is not None:
            leverage = vol_scale * (df_pnl_all['value'].std() * np.sqrt(261))**(-1)
            df_pnl['value'] = df_pnl['value'] * leverage

        pnn = ('PNL_' + sig + neg) if pnl_name is None else pnl_name
        # Populating the signal dictionary is required for the display methods:
        self.signal_df[pnn] = dfw.loc[:, ['cid', 'real_date', 'sig']]

        df_pnl['xcat'] = pnn
        if pnn in self.pnl_names:
            self.df = self.df[~(self.df['xcat'] == pnn)]
        else:
            self.pnl_names = self.pnl_names + [pnn]

        agg_df = pd.concat([self.df, df_pnl[self.df.columns]])
        self.df = agg_df.reset_index(drop=True)

    def make_long_pnl(self, vol_scale: Optional[float] = None, 
                      label: Optional[str] = None):
        """
        The long-only returns will be computed which act as a basis for comparison
        against the signal-adjusted returns. Will take a long-only position in the
        category passed to the parameter 'self.ret'.

        :param <bool> vol_scale: ex-post scaling of PnL to annualized volatility given.
            This is for comparative visualization and not out-of-sample, and is applied
            to the long-only position. Default is None.
        :param <str> label: associated label that will be mapped to the long-only
            DataFrame. The label will be used in the plotting graphic for plot_pnls().
            If a label is not defined, the default will be the name of the return
            category.
        """
        
        if vol_scale is not None:
            if not isinstance(vol_scale, (float, int)):
                raise TypeError("The volatility scale `vol_scale`"
                                "must be a numerical value.")

        if label is None:
            label = self.ret

        dfx = self.df[self.df['xcat'].isin([self.ret])]

        df_long = self.long_only_pnl(dfw=dfx, vol_scale=vol_scale,
                                     label=label)

        self.df = pd.concat([self.df, df_long])

        if label not in self.pnl_names:
            self.pnl_names = self.pnl_names + [label]

        self.df = self.df.reset_index(drop=True)

    @staticmethod
    def long_only_pnl(dfw: pd.DataFrame, vol_scale: float = None, label: str = None):
        """
        Method used to compute the PnL accrued from simply taking a long-only position in
        the category, 'self.ret'. The returns from the category are not predicated on any
        exogenous signal.

        :param <pd.DataFrame> dfw:
        :param <bool> vol_scale: ex-post scaling of PnL to annualized volatility given.
            This is for comparative visualization and not out-of-sample. Default is none.
        :param <str> label: associated label that will be mapped to the long-only
            DataFrame.

        :return <pd.DataFrame> panel_pnl: standardised dataframe containing exclusively
            the return category, and the long-only panel return.
        """

        dfw_long = dfw.reset_index(drop=True)

        panel_pnl = dfw_long.groupby(['real_date']).sum(numeric_only=True)
        panel_pnl = panel_pnl.reset_index(level=0)
        panel_pnl['cid'] = 'ALL'
        panel_pnl['xcat'] = label

        if vol_scale:
            leverage = vol_scale * (panel_pnl['value'].std() * np.sqrt(261)) ** (-1)
            panel_pnl['value'] = panel_pnl['value'] * leverage

        return panel_pnl[['cid', 'xcat', 'real_date', 'value']]

    def plot_pnls(self, pnl_cats: List[str] = None, pnl_cids: List[str] = ['ALL'],
                  start: str = None, end: str = None,
                  facet: bool = False,
                  ncol: int = 3, same_y: bool = True,
                  title: str = "Cumulative Naive PnL", xcat_labels: List[str] = None,
                  xlab: str = "",
                  ylab: str = "% of risk capital, no compounding",
                  share_axis_labels: bool = True,
                  figsize: Tuple = (12, 7), aspect: float = 1.7,
                  height: float = 3, label_adj: float = 0.05,
                  title_adj: float = 0.95, y_label_adj: float = 0.95,) -> None:
        """
        Plot line chart of cumulative PnLs, single PnL, multiple PnL types per
        cross section, or multiple cross sections per PnL type.

        :param <List[str]> pnl_cats: list of PnL categories that should be plotted.
        :param <List[str]> pnl_cids: list of cross sections to be plotted;
            default is 'ALL' (global PnL).
            Note: one can only have multiple PnL categories or multiple cross sections,
            not both.
        :param <str> start: earliest date in ISO format. Default is None and earliest
            date in df is used.
        :param <str> end: latest date in ISO format. Default is None and latest date
            in df is used.
        :param <bool> facet: parameter to control whether each PnL series is plotted on
            its own respective grid using Seaborn's FacetGrid. Default is False and all
            series will be plotted in the same graph.
        :param <int> ncol: number of columns in facet grid. Default is 3. If the total
            number of PnLs is less than ncol, the number of columns will be adjusted on
            runtime.
        :param <bool> same_y: if True (default) all plots in facet grid share same y axis.
        :param <str> title: allows entering text for a custom chart header.
        :param <List[str]> xcat_labels: custom labels to be used for the PnLs.
        :param <str> xlab: label for x-axis of the plot (or subplots if faceted),
            default is None (empty string)..
        :param <str> ylab: label for y-axis of the plot (or subplots if faceted),
            default is '% of risk capital, no compounding'.
        :param <bool> share_axis_labels: if True (default) the axis labels are shared by
            all subplots in the facet grid.
        :param <tuple> figsize: tuple of plot width and height. Default is (12 , 7).
        :param <float> aspect: width-height ratio for plots in facet. Default is 1.7.
        :param <float> height: height of plots in facet. Default is 3.
        :param <float> label_adj: parameter that sets bottom of figure to fit the label.
            Default is 0.05.
        :param <float> title_adj: parameter that sets top of figure to accommodate title.
            Default is 0.95.
        :param <float> y_label_adj: parameter that sets left of figure to fit the y-label.
        """

        if pnl_cats is None:
            pnl_cats = self.pnl_names
        else:
            pnl_cats_error = f"List of PnL categories expected - received " \
                             f"{type(pnl_cats)}."
            assert isinstance(pnl_cats, list), pnl_cats_error

            pnl_cats_copy = pnl_cats.copy()
            pnl_cats = [pnl for pnl in pnl_cats if pnl in self.pnl_names]

            dif = set(pnl_cats_copy).difference(set(pnl_cats))
            if dif:
                print(f"The PnL(s) requested, {dif}, have not been defined on the "
                      f"Class. The defined PnL(s) are {self.pnl_names}.")
            elif len(pnl_cats) == 0:
                raise ValueError("There are not any valid PnL(s) to display given the "
                                 "request.")

        error_message = "Either pnl_cats or pnl_cids must be a list of length 1"
        assert (len(pnl_cats) == 1) | (len(pnl_cids) == 1), error_message

        # adjust ncols of the facetgrid if necessary
        if max([len(pnl_cats), len(pnl_cids)]) < ncol:
            ncol = max([len(pnl_cats), len(pnl_cids)])

        dfx = reduce_df(
            self.df, pnl_cats, pnl_cids, start, end, self.black, out_all=False
        )

        if max([len(pnl_cats), len(pnl_cids)]) < ncol:
            ncol = max([len(pnl_cats), len(pnl_cids)])

        error_message = "The number of custom labels must match the defined number of " \
                        "categories in pnl_cats."
        if xcat_labels is not None:
            assert(len(xcat_labels) == len(pnl_cats)), error_message

        else:
            xcat_labels = pnl_cats.copy()

        no_cids = len(pnl_cids)

        sns.set_theme(
            style='whitegrid', palette='colorblind', rc={'figure.figsize': figsize}
        )

        if no_cids == 1:
            plot_by = "xcat"
            col_order = pnl_cats
            labels = xcat_labels
            legend_title = "PnL Category(s)"
        else:
            plot_by = "cid"
            col_order = pnl_cids
            if xcat_labels is not None:
                labels = pnl_cids.copy()
            legend_title = "Cross Section(s)"

        dfx['cum_value'] = dfx.groupby(plot_by).cumsum(numeric_only=True)

        if facet:
            fg = sns.FacetGrid(
                data=dfx, col=plot_by, col_wrap=ncol, sharey=same_y, aspect=aspect,
                height=height, col_order=col_order, legend_out=True
            )
            fg.fig.suptitle(title, fontsize=20,)

            fg.fig.subplots_adjust(top=title_adj, bottom=label_adj, left=y_label_adj)

            fg.map_dataframe(
                sns.lineplot, x="real_date", y="cum_value", hue=plot_by,
                hue_order=col_order, estimator=None, lw=1
            )
            for ix, ax in enumerate(fg.axes.flat):
                ax.axhline(y=0, color="black", linestyle='--', linewidth=1)
                if no_cids == 1:
                    ax.set_title(xcat_labels[ix])
            
            if no_cids > 1:
                fg.set_titles(row_template='', col_template='{col_name}')
            
            if share_axis_labels:
                fg.set_axis_labels("", "")
                fg.fig.supxlabel(xlab)
                fg.fig.supylabel(ylab)
            else:
                fg.set_axis_labels(xlab, ylab)
            
        else:
            fg = sns.lineplot(data=dfx, x="real_date", y="cum_value", hue=plot_by, 
                                hue_order=col_order, estimator=None, lw=1)
            leg = fg.axes.get_legend()
            plt.title(title, fontsize=20)
            plt.legend(labels=labels, title=legend_title,)
            plt.xlabel(xlab)
            plt.ylabel(ylab)

        if no_cids == 1:
            if facet:
                labels = labels[::-1]
        else:
            labels = labels[::-1]
        
        plt.axhline(y=0, color='black', linestyle='--', lw=1)
        plt.show()

    def signal_heatmap(self, pnl_name: str, pnl_cids: List[str] = None,
                       start: str = None, end: str = None, freq: str = 'm',
                       title: str = "Average applied signal values",
                       x_label: str = "", y_label: str = "",
                       figsize: Optional[Tuple[float, float]] = None):

        """
        Display heatmap of signals across times and cross-sections.

        :param <str> pnl_name: name of naive PnL whose signals are displayed.
            N.B.: Signal is here is the value that actually determines
            the concurrent PnL.
        :param <List[str]> pnl_cids: cross-sections. Default is all available.
        :param <str> start: earliest date in ISO format. Default is None and earliest
            date in df is used.
        :param <str> end: latest date in ISO format. Default is None and latest date
            in df is used.
        :param <str> freq: frequency for which signal average is displayed.
            Default is monthly ('m'). The only alternative is quarterly ('q').
        :param <str> title: allows entering text for a custom chart header.
        :param <str> x_label: label for the x-axis. Default is None.
        :param <str> y_label: label for the y-axis. Default is None.
        :param <(float, float)> figsize: width and height in inches.
            Default is (14, number of cross sections).
        """

        assert isinstance(pnl_name, str), "The method expects to receive a single " \
                                          "PnL name."
        error_cats = f"The PnL passed to 'pnl_name' parameter is not defined. The " \
                     f"possible options are {self.pnl_names}."
        assert pnl_name in self.pnl_names, error_cats

        error_time = "Defined time-period must be monthly ('m') or quarterly ('q')"
        assert isinstance(freq, str) and freq in ['m', 'q'], error_time

        error_cids = f"Cross-sections not available. Available cids are:" \
                     f"{self.cids}."

        if pnl_cids is None:
            pnl_cids = self.cids
        else:
            assert set(pnl_cids) <= set(self.cids), error_cids

        assert isinstance(x_label, str), f"<str> expected - received {type(x_label)}."
        assert isinstance(y_label, str), f"<str> expected - received {type(y_label)}."

        dfx = self.signal_df[pnl_name]
        dfw = dfx.pivot(index='real_date', columns='cid', values='sig')
        dfw = dfw[pnl_cids]

        if start is None:
            start = dfw.index[0]
        elif end is None:
            end = dfw.index[-1]

        dfw = dfw.truncate(before=start, after=end)

        dfw = dfw.resample(freq, axis=0).mean()

        if figsize is None:
            figsize = (14, len(pnl_cids))

        fig, ax = plt.subplots(figsize=figsize)
        dfw = dfw.transpose()
        dfw.columns = [str(d.strftime('%d-%m-%Y')) for d in dfw.columns]
        sns.heatmap(dfw, cmap="vlag_r", center=0)

        ax.set(xlabel=x_label, ylabel=y_label)
        ax.set_yticklabels(ax.get_yticklabels(), rotation=0)
        ax.set_title(title, fontsize=14)

        plt.show()

    def agg_signal_bars(self, pnl_name: str, freq: str = 'm',
                        metric: str = "direction", title: str = None,
                        y_label: str = "Sum of Std. across the Panel"):
        """
        Display aggregate signal strength and - potentially - direction.

        :param <str> pnl_name: name of the PnL whose signal is to be visualized.
            N.B.: The referenced signal corresponds to the series that determines the
            concurrent PnL.
        :param <str> freq: frequency at which the signal is visualized. Default is
            monthly ('m'). The alternative is quarterly ('q').
        :param <str> metric: the type of signal value. Default is "direction".
            Alternative is "strength".
        :param <str> title: allows entering text for a custom chart header. Default will
            be "Directional Bar Chart of <pnl_name>.".
        :param <str> y_label: label for the y-axis. Default is the sum of standard
            deviations across the panel corresponding to the default signal
            transformation: 'zn_score_pan'.

        """

        assert isinstance(pnl_name, str), "The method expects to receive a single " \
                                          "PnL name."
        error_cats = f"The PnL passed to 'pnl_name' parameter is not defined. The " \
                     f"possible options are {self.pnl_names}."
        assert pnl_name in self.pnl_names, error_cats

        error_time = "Defined time-period must either be monthly, m, or quarterly, q."
        assert isinstance(freq, str) and freq in ['m', 'q'], error_time

        metric_error = "The metric must either be 'direction' or 'strength'."
        assert metric in ['direction', "strength"], metric_error

        if title is None:
            title = f"Directional Bar Chart of {pnl_name}."

        dfx = self.signal_df[pnl_name]
        dfw = dfx.pivot(index='real_date', columns='cid', values='sig')

        # The method allows for visually understanding the overall direction of the
        # aggregate signal but also gaining insight into the proportional exposure to the
        # respective signal by measuring the absolute value, the size of the signal.
        # Is the PnL value generated by large returns or a large signal, position ?
        if metric == "strength":
            dfw = dfw.abs()

        dfw = dfw.resample(freq, axis=0).mean()
        # Sum across the timestamps to compute the aggregate signal according to the
        # down-sampling frequency.
        df_s = dfw.sum(axis=1)
        index = np.array(df_s.index)
        df_signal = pd.DataFrame(data=df_s.to_numpy(), columns=["aggregate_signal"],
                                 index=index)

        df_signal = df_signal.reset_index(level=0)
        df_signal = df_signal.rename({"index": ""}, axis="columns")
        dates = [pd.Timestamp(d) for d in df_signal['']]
        df_signal[''] = np.array(dates)

        plt.style.use("ggplot")

        fig, ax = plt.subplots()
        df_signal.plot.bar(x="", y="aggregate_signal", ax=ax, title=title,
                           ylabel=y_label, legend=False)

        ticklabels = [''] * len(df_signal)
        skip = len(df_signal) // 12
        ticklabels[::skip] = df_signal[''].iloc[::skip].dt.strftime('%Y-%m-%d')
        ax.xaxis.set_major_formatter(mticker.FixedFormatter(ticklabels))

        fig.autofmt_xdate()

        def fmt(x, pos=0, max_i=len(ticklabels) - 1):
            i = int(x)
            i = 0 if i < 0 else max_i if i > max_i else i
            return dates[i]

        ax.fmt_xdata = fmt
        plt.show()

    def evaluate_pnls(self, pnl_cats: List[str], pnl_cids: List[str] = ['ALL'],
                      start: str = None, end: str = None):

        """
        Table of key PnL statistics.

        :param <List[str]> pnl_cats: list of PnL categories that should be plotted.
        :param <List[str]> pnl_cids: list of cross-sections to be plotted; default is
            'ALL' (global PnL).
            Note: one can only have multiple PnL categories or multiple cross-sections,
            not both.
        :param <str> start: earliest date in ISO format. Default is None and earliest
            date in df is used.
        :param <str> end: latest date in ISO format. Default is None and latest date
            in df is used.

        :return <pd.DataFrame>: standardized DataFrame with key PnL performance
            statistics.
        """

        error_cids = "List of cross-sections expected."
        error_xcats = "List of PnL categories expected."
        assert isinstance(pnl_cids, list), error_cids
        assert isinstance(pnl_cats, list), error_xcats
        assert all([isinstance(elem, str) for elem in pnl_cids]), error_cids
        assert all([isinstance(elem, str) for elem in pnl_cats]), error_xcats

        if pnl_cats is None:
            # The field, self.pnl_names, is a data structure that stores the name of the
            # category assigned to PnL values. Each time make_pnl() method is called, the
            # computed DataFrame will have an associated category established by the
            # logical method: ('PNL_' + sig) if pnl_name is None else pnl_name. Each
            # category will be held in the data structure.
            pnl_cats = self.pnl_names
        else:
            if not set(pnl_cats) <= set(self.pnl_names):
                missing = [pnl for pnl in pnl_cats if pnl not in self.pnl_names]
                pnl_error = f"Received PnL categories have not been defined. The PnL " \
                            f"category(s) which has not been defined is: {missing}. " \
                            f"The produced PnL category(s) are {self.pnl_names}."
                raise ValueError(pnl_error)

        assert (len(pnl_cats) == 1) | (len(pnl_cids) == 1)

        dfx = reduce_df(self.df, pnl_cats, pnl_cids, start,
                        end, self.black, out_all=False)

        groups = "xcat" if len(pnl_cids) == 1 else "cid"
        stats = ['Return (pct ar)', 'St. Dev. (pct ar)', 'Sharpe Ratio', 'Sortino Ratio',
                 'Max 21-day draw', 'Max 6-month draw', 'Traded Months']

        # If benchmark tickers have been passed into the Class and if the tickers are
        # present in self.dfd.
        list_for_dfbm = []

        if self.bm_bool and bool(self._bm_dict):

            list_for_dfbm = list(self._bm_dict.keys())
            for bm in list_for_dfbm:
                stats.insert(len(stats) - 1, f"{bm} correl")

        dfw = dfx.pivot(index='real_date', columns=groups, values='value')
        df = pd.DataFrame(columns=dfw.columns, index=stats)

        df.iloc[0, :] = dfw.mean(axis=0) * 261
        df.iloc[1, :] = dfw.std(axis=0) * np.sqrt(261)
        df.iloc[2, :] = df.iloc[0, :] / df.iloc[1, :]
        dsd = dfw.apply(lambda x: np.sqrt(np.sum(x[x < 0]**2)/len(x))) * np.sqrt(261)
        df.iloc[3, :] = df.iloc[0, :] / dsd
        df.iloc[4, :] = dfw.rolling(21).sum().min()
        df.iloc[5, :] = dfw.rolling(6*21).sum().min()
        if len(list_for_dfbm) > 0:
            bm_df = pd.concat(
                list(self._bm_dict.values()),
                axis=1
            )
            for i, bm in enumerate(list_for_dfbm):
                index = dfw.index.intersection(bm_df.index)
                correlation = dfw.loc[index].corrwith(
                    bm_df.loc[index].iloc[:, i],
                    axis=0,
                    method='pearson',
                    drop=True
                )
                df.iloc[6 + i, :] = correlation

        df.iloc[6 + len(list_for_dfbm), :] = dfw.resample('M').sum().count()

        return df

    def print_pnl_names(self):
        """
        Print list of names of available PnLs in the class instance.
        """

        print(self.pnl_names)

    def pnl_df(self, pnl_names: List[str] = None,  cs: bool = False):
        """
        Return dataframe with PnLs.

        :param <List[str]> pnl_names: list of names of PnLs to be returned.
            Default is 'ALL'.
        :param <bool> cs: inclusion of cross section PnLs. Default is False.

        :return custom DataFrame with PnLs
        """
        selected_pnls = pnl_names if pnl_names is not None else self.pnl_names

        filter_1 = self.df['xcat'].isin(selected_pnls)
        filter_2 = self.df['cid'] == 'ALL' if not cs else True

        return self.df[filter_1 & filter_2]


if __name__ == "__main__":
    cids = ['AUD', 'CAD', 'GBP', 'NZD', 'USD', 'EUR']
    xcats = ['EQXR_NSA', 'CRY', 'GROWTH', 'INFL', 'DUXR']

    cols_1 = ['earliest', 'latest', 'mean_add', 'sd_mult']
    df_cids = pd.DataFrame(index=cids, columns=cols_1)
    df_cids.loc['AUD', :] = ['2008-01-03', '2020-12-31', 0.5, 2]
    df_cids.loc['CAD', :] = ['2010-01-03', '2020-11-30', 0, 1]
    df_cids.loc['GBP', :] = ['2012-01-03', '2020-11-30', -0.2, 0.5]
    df_cids.loc['NZD'] = ['2002-01-03', '2020-09-30', -0.1, 2]
    df_cids.loc['USD'] = ['2015-01-03', '2020-12-31', 0.2, 2]
    df_cids.loc['EUR'] = ['2008-01-03', '2020-12-31', 0.1, 2]

    cols_2 = cols_1 + ['ar_coef', 'back_coef']

    df_xcats = pd.DataFrame(index=xcats, columns=cols_2)
    df_xcats.loc['EQXR_NSA'] = ['2000-01-03', '2020-12-31', 0.1, 1, 0, 0.3]
    df_xcats.loc['CRY'] = ['2000-01-01', '2020-10-30', 1, 2, 0.95, 1]
    df_xcats.loc['GROWTH'] = ['2010-01-03', '2020-10-30', 1, 2, 0.9, 1]
    df_xcats.loc['INFL'] = ['2001-01-01', '2020-10-30', 1, 2, 0.8, 0.5]
    df_xcats.loc['DUXR'] = ['2000-01-01', '2020-12-31', 0.1, 0.5, 0, 0.1]

    black = {'AUD': ['2006-01-01', '2015-12-31'], 'GBP': ['2022-01-01', '2100-01-01']}
    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    # Instantiate a new instance to test the long-only functionality.
    # Benchmarks are used to calculate correlation against PnL series.
    pnl = NaivePnL(
        dfd, ret="EQXR_NSA", sigs=["CRY", "GROWTH", "INFL"], cids=cids,
        start="2000-01-01", blacklist=black, bms=["EUR_EQXR_NSA", "USD_EQXR_NSA"]
    )

    pnl.make_pnl(
        sig="GROWTH", sig_op="zn_score_pan", sig_neg=True, rebal_freq="monthly",
        vol_scale=5, rebal_slip=1, min_obs=250, thresh=2
    )

    pnl.make_long_pnl(vol_scale=10, label="Long")

    df_eval = pnl.evaluate_pnls(
        pnl_cats=["PNL_GROWTH_NEG"], start="2015-01-01", end="2020-12-31"
    )

    pnl.agg_signal_bars(
        pnl_name="PNL_GROWTH_NEG", freq="m", metric="direction", title=None,
    )
    pnl.plot_pnls(
        pnl_cats=["PNL_GROWTH_NEG", "Long"], facet=False,
        xlab="date", ylab="%",
        )
    pnl.plot_pnls(
        pnl_cats=["PNL_GROWTH_NEG", "Long"], facet=False, xcat_labels=["S_1", "S_2"],
        xlab="date", ylab="%"
    )
    pnl.plot_pnls(
        pnl_cats=["PNL_GROWTH_NEG", "Long"], facet=True, xcat_labels=["S_1", "S_2"]
    )    
    pnl.plot_pnls(
        pnl_cats=["PNL_GROWTH_NEG", "Long"], facet=True,
    )
    
    pnl.plot_pnls(  pnl_cats=["PNL_GROWTH_NEG"], 
                    pnl_cids=cids,
                    xcat_labels=None
                    )

    pnl.plot_pnls(  pnl_cats=["PNL_GROWTH_NEG"], 
                    pnl_cids=cids,
                    facet=True,
                    xcat_labels=None
                    )
    
    pnl.plot_pnls(  pnl_cats=["PNL_GROWTH_NEG"],
                    pnl_cids=cids,
                    same_y=True,
                    facet=True,
                    xcat_labels=None,
                    share_axis_labels=False,
                    xlab='Date',
                    ylab='PnL',
                    y_label_adj=0.1,)




import numpy as np
import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
from typing import List, Union, Tuple, Optional

from macrosynergy.management.simulate_quantamental_data import make_qdf
from macrosynergy.management.check_availability import reduce_df


def view_timelines(
    df: pd.DataFrame,
    xcats: Optional[List[str]] = None,
    cids: Optional[List[str]] = None,
    intersect: bool = False,
    val: str = "value",
    cumsum: bool = False,
    start: str = "2000-01-01",
    end: Optional[str] = None,
    ncol: int = 3,
    legend_ncol: int = 1,
    same_y: bool = True,
    all_xticks: bool = False,
    xcat_grid: bool = False,
    xcat_labels: Optional[List[str]] = None,
    single_chart: bool = False,
    label_adj: float = 0.05,
    title: Optional[str] = None,
    title_adj: float = 0.95,
    title_xadj: float = 0.5,
    title_fontsize: int = 16,
    cs_mean: bool = False,
    size: Tuple[float, float] = (12, 7),
    aspect: float = 1.7,
    height: float = 3.0,
    legend_fontsize: int = 12,
):

    """Displays a facet grid of time line charts of one or more categories.

    :param <pd.Dataframe> df: standardized DataFrame with the necessary columns:
        'cid', 'xcats', 'real_date' and at least one column with values of interest.
    :param <List[str]> xcats: extended categories to plot. Default is all in DataFrame.
    :param <List[str]> cids: cross sections to plot. Default is all in DataFrame.
        If this contains only one cross section a single line chart is created.
    :param <bool> intersect: if True only retains cids that are available for all xcats.
        Default is False.
    :param <str> val: name of column that contains the values of interest.
        Default is 'value'.
    :param <bool> cumsum: plot cumulative sum of the values over time. Default is False.
    :param <str> start: earliest date in ISO format. Default is earliest date available.
    :param <str> end: latest date in ISO format. Default is latest date available.
    :param <int> ncol: number of columns in facet grid. Default is 3.
    :param <int> legend_ncol: number of columns in legend. Default is 1.
    :param <bool> same_y: if True (default) all plots in facet grid share same y axis.
    :param <bool> all_xticks:  if True x-axis tick labels are added to all plots in grid.
        Default is False, i.e only the lowest row displays the labels.
    :param <bool> xcat_grid: if True, shows a facet grid of line charts for each xcat
        for a single cross section. Default is False, only one cross section is allowed
        with this option.
    :param <List[str]> xcat_labels: labels to be used for xcats. If not defined, the
        labels will be identical to extended categories.
    :param <bool> single_chart: if True, all lines are plotted in a single chart.
    :param <str> title: chart heading. Default is no title.
    :param <float> title_adj: parameter that sets top of figure to accommodate title.
        Default is 0.95.
    :param <float> title_xadj: parameter that sets x position of title. Default is 0.5.
    :param <int> title_fontsize: font size of title. Default is 16.
    :param <float> label_adj: parameter that sets bottom of figure to fit the label.
        Default is 0.05.
    :param <bool> cs_mean: if True this adds a line of cross-sectional averages to
        the line charts. This is only allowed for function calls with a single
        category. Default is False.
    :param <Tuple[float]> size: two-element tuple setting width/height of single cross
        section plot. Default is (12, 7). This is irrelevant for facet grid.
    :param <float> aspect: width-height ratio for plots in facet. Default is 1.7.
    :param <float> height: height of plots in facet. Default is 3.
    :param <int> legend_fontsize: font size of legend. Default is 12.

    """

    if not set(df.columns).issuperset({"cid", "xcat", "real_date", val}):
        fail_str: str = (
            f"Error : Tried to standardize DataFrame but failed."
            f"DataFrame not in the correct format. Please ensure "
            f"that the DataFrame has the following columns: "
            f"'cid', 'xcat', 'real_date' and column='{val}'."
        )
        try:
            dft = df.reset_index()
            assert set(dft.columns).issuperset({"cid", "xcat", "real_date", val})

            df = dft.copy()
        except Exception as e:
            raise ValueError(fail_str)

    df["real_date"] = pd.to_datetime(df["real_date"], format="%Y-%m-%d")

    if not isinstance(cs_mean, bool):
        raise TypeError(f"`cs_mean` parameter must be a Boolean object.")

    if isinstance(xcats, str):
        xcats = [xcats]
    if isinstance(cids, str):
        cids = [cids]

    for varx, namex in zip([xcats, cids], ["xcats", "cids"]):
        if (
            (not isinstance(varx, list))
            or (not all(isinstance(x, str) for x in varx))
            or (not varx)
        ):
            raise TypeError(f"`{namex}` must be a list of strings.")

    if cs_mean and len(xcats) > 1:
        raise ValueError(
            f"cs_mean can only be set to True "
            "if a single category is passed. The "
            f"received categories are {xcats}."
        )

    if not isinstance(xcat_grid, bool):
        raise TypeError("`xcat_grid` parameter must be a Boolean object.")

    if not isinstance(single_chart, bool):
        raise TypeError("`single_chart` parameter must be a Boolean object.")
    if xcat_grid and single_chart:
        raise ValueError("xcat_grid and single_chart cannot both be True.")

    df, xcats, cids = reduce_df(
        df, xcats, cids, start, end, out_all=True, intersect=intersect
    )

    # NOTE: casting var(cids) to list if it is a string is dependent on the reduce_df function

    if xcat_grid:

        if not len(cids) == 1:
            raise ValueError(
                "`xcat_grid` can only be set to True if a "
                "single cross-section (`cids`) is passed."
            )

    if cumsum:
        df[val] = (
            df.sort_values(["cid", "xcat", "real_date"])[["cid", "xcat", val]]
            .groupby(["cid", "xcat"])
            .cumsum()
        )

    max_plots: int = 1
    if (len(cids) == 1) and xcat_grid:
        max_plots = len(xcats)
    elif (len(cids) > 1) and not single_chart:
        max_plots = len(cids)
    ncol: int = min(ncol, max_plots)

    sns.set(style="darkgrid")
    cross_mean: Optional[pd.DataFrame] = None
    cs_label: str = f"cross-sectional average of {xcats[0]}."
    if cs_mean and (len(cids) > 1):
        dfw: pd.DataFrame = df.pivot(index="real_date", columns="cid", values="value")
        cm_series: pd.Series = dfw.mean(axis=1)
        cross_mean: pd.DataFrame = pd.DataFrame(
            data=cm_series.to_numpy(), index=cm_series.index, columns=["average"]
        ).reset_index(level=0)

        if xcat_labels is not None:
            if len(xcat_labels) == (len(xcats) + 1):
                cs_label = xcat_labels.pop(-1)

    # Replace xcats in the DF with their corresponding labels
    if xcat_labels is not None:
        for xc, xl in zip(xcats, xcat_labels):
            df["xcat"] = df["xcat"].replace(xc, xl)
        xcats = xcat_labels

    sns.set(rc={"figure.figsize": size})
    plt.rcParams["figure.figsize"] = size

    fg: Optional[sns.FacetGrid] = None
    ax: Optional[plt.Axes] = None
    if len(cids) == 1:
        if xcat_grid:
            fg: sns.FacetGrid = sns.FacetGrid(
                df,
                col="xcat",
                col_wrap=ncol,
                sharey=same_y,
                height=height,
                aspect=aspect,
                col_order=xcats,
            )
            fg.map_dataframe(
                sns.lineplot,
                x="real_date",
                y=val,
                hue="xcat",
                hue_order=xcats,
                estimator=None,
            )

            fg.map(plt.axhline, y=0, c=".5")
            fg.set_axis_labels("", "")
            fg.set_titles("{col_name}")

        else:
            ax: plt.Axes = plt.gca()
            for xc in xcats:
                dfc: pd.DataFrame = df[df["xcat"] == xc]
                ax.plot(dfc["real_date"], dfc[val], label=xc)

            plt.axhline(y=0, c=".5")

            ax.set_xlabel("")
            ax.set_ylabel("")
            ax.legend(
                ncol=legend_ncol,
                fontsize=legend_fontsize,
            )

    else:
        if not single_chart:
            fg: sns.FacetGrid = sns.FacetGrid(
                df,
                col="cid",
                col_wrap=ncol,
                sharey=same_y,
                height=height,
                aspect=aspect,
                col_order=cids,
            )
            fg.map_dataframe(
                sns.lineplot,
                x="real_date",
                y=val,
                hue="xcat",
                hue_order=xcats,
                estimator=None,
            )

            if cs_mean:
                fg.map(
                    sns.lineplot,
                    x="real_date",
                    y="average",
                    color="red",
                    estimator=None,
                    label=cs_label,
                    data=cross_mean,
                )

            fg.map(plt.axhline, y=0, c=".5")
            fg.set_titles(col_template="{col_name}")
            fg.set_axis_labels("", "")
            if cs_mean or (len(xcats) > 1):
                fg.add_legend(
                    loc="lower center", ncol=legend_ncol, fontsize=legend_fontsize
                )

        else:
            ax: plt.Axes = sns.lineplot(
                data=df,
                x="real_date",
                y=val,
                hue="cid",
                hue_order=cids,
                estimator=None,
            )

            if cs_mean:
                ax: plt.Axes = sns.lineplot(
                    data=cross_mean,
                    x="real_date",
                    y="average",
                    color="red",
                    estimator=None,
                    label=cs_label,
                )

            plt.axhline(y=0, c=".5")
            ax.set_xlabel("")
            ax.set_ylabel("")
            ax.legend(ncol=legend_ncol, fontsize=legend_fontsize)


    if all_xticks:
        if fg is not None:
            for ax in fg.axes.flat:
                ax.tick_params(labelbottom=True, pad=0)
        else:
            ax.tick_params(labelbottom=True, pad=0)

    if fg is not None:
        fg.figure.subplots_adjust(bottom=label_adj)
    else:
        plt.subplots_adjust(bottom=label_adj)
        
    if title is not None:
        if fg is not None:
            fg.figure.suptitle(
                title,
                y=title_adj,
                fontsize=title_fontsize,
                x=0.5, horizontalalignment="center"
            )
        else:
            ax.set_title(title, y=title_adj, fontsize=title_fontsize,
                            x=0.5, horizontalalignment="center")

    plt.show()


if __name__ == "__main__":

    cids = ["AUD", "CAD", "GBP", "NZD"]
    xcats = ["XR", "CRY", "INFL"]
    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc[
        "AUD",
    ] = ["2010-01-01", "2020-12-31", 0.2, 0.2]
    df_cids.loc[
        "CAD",
    ] = ["2011-01-01", "2020-11-30", 0, 1]
    df_cids.loc[
        "GBP",
    ] = ["2012-01-01", "2020-11-30", 0, 2]
    df_cids.loc[
        "NZD",
    ] = ["2012-01-01", "2020-09-30", -0.1, 3]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )

    df_xcats.loc[
        "XR",
    ] = ["2010-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc[
        "INFL",
    ] = ["2015-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc[
        "CRY",
    ] = ["2013-01-01", "2020-10-30", 1, 2, 0.95, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfdx = dfd[~((dfd["cid"] == "AUD") & (dfd["xcat"] == "XR"))]

    view_timelines(
        dfd,
        xcats=["XR", "CRY"],
        cids=cids[0],
        size=(10, 5),
        title="AUD Return and Carry",
        label_adj=0.1,
    )

    view_timelines(
        dfd,
        xcats=["XR", "CRY", "INFL"],
        cids=cids[0],
        xcat_grid=True,
        title_adj=0.8,
        xcat_labels=["Return", "Carry", "Inflation"],
        title="AUD Return, Carry & Inflation",
    )

    view_timelines(dfd, xcats=["CRY"], cids=cids, ncol=2, title="Carry", cs_mean=True)
    view_timelines(
        dfd,
        xcats=["CRY"],
        cids=cids,
        ncol=2,
        title="Carry",
        cs_mean=True,
        xcat_labels=["Carry", "cs-mean-1"],
    )

    view_timelines(
        dfd, xcats=["XR"], cids=cids, ncol=2, cumsum=True, same_y=False, aspect=2
    )

    dfd = dfd.set_index("real_date")
    view_timelines(
        dfd,
        xcats=["XR"],
        cids=cids,
        ncol=2,
        cumsum=True,
        same_y=False,
        aspect=2,
        single_chart=True,
    )

    view_timelines(
        dfd,
        xcats=["XR"],
        single_chart=True,
        cids=cids,
        start="2010-01-01",
        title="AUD Return Comparison",
        cs_mean=True,
        aspect=1.5,
        title_adj=0.92,
        label_adj=0.08,
        xcat_labels=["AUD Return", "cs-mean"],
    )



import numpy as np
import pandas as pd
from typing import *
import warnings

from macrosynergy.management.shape_dfs import reduce_df
from macrosynergy.management.simulate_quantamental_data import make_qdf, make_test_df
from macrosynergy.management.utils import is_valid_iso_date


def linear_composite_on_cid(
    df: pd.DataFrame,
    xcats: Union[str, List[str]],
    cids: Optional[List[str]] = None,
    weights: Union[List[float], np.ndarray, pd.Series] = None,
    signs: Union[List[float], np.ndarray, pd.Series] = None,
    start: str = None,
    end: str = None,
    normalize_weights: bool = False,
    complete_xcats: bool = True,
    new_xcat="NEW",
):

    if not len(xcats) == len(weights) == len(signs):
        raise ValueError("xcats, weights, and signs must have same length")
    if not np.isclose(np.sum(weights), 1) or normalize_weights:
        if not normalize_weights:
            warnings.warn("`weights` does not sum to 1 and will be normalized. w←w/∑w")
        weights = weights / np.sum(weights)
    if not np.all(np.isin(signs, [1, -1])):
        warnings.warn("signs must be 1 or -1. They will be coerced to 1 or -1.")
        signs = np.abs(signs) / signs  # should be faster?

    # main function is here and below.
    weights = pd.Series(weights * signs, index=xcats)

    dfc: pd.DataFrame = reduce_df(df=df, cids=cids, xcats=xcats, start=start, end=end)

    # dataframe with the xcats as columns and rows as cid-date combinations
    dfc_wide = dfc.set_index(["cid", "real_date", "xcat"])["value"].unstack(level=2)
    # dataframe for weights with same index as dfc_wide: each column will be a weight
    weights_wide = pd.DataFrame(
        data=[weights.sort_index()], index=dfc_wide.index, columns=dfc_wide.columns
    )
    # boolean mask to help us work out the calcs
    mask = dfc_wide.isna()
    # series with an index of dfc_wide, and a value equal to the sum of the weights
    weights_sum = weights_wide[~mask].abs().sum(axis=1)
    # re-weighting the weights to sum to 1 considering the available xcats
    adj_weights_wide = weights_wide[~mask].div(weights_sum, axis=0)
    # final single series: the linear combination of the xcats and the weights

    out_df = (dfc_wide * adj_weights_wide).sum(axis=1)

    if complete_xcats:
        out_df[mask.any(axis=1)] = np.NaN
    else:
        out_df[mask.all(axis=1)] = np.NaN

    out_df = out_df.reset_index().rename(columns={0: "value"})
    out_df["xcat"] = new_xcat
    out_df = out_df[["cid", "xcat", "real_date", "value"]]

    return out_df


def linear_composite_on_xcat(
    df: pd.DataFrame,
    xcat: str,
    cids: List[str],
    weights: str,
    normalize_weights: bool = False,
    complete_cids: bool = False,
    update_freq: str = "M",
    new_cid="GLB",
):
    # sort and filter
    df = df.copy().sort_values(by=["cid", "xcat", "real_date"])
    cids_mask = df["cid"].isin(cids)

    # seekect the target and weights
    target_df: pd.DataFrame = df[(df["xcat"] == xcat) & cids_mask].copy()
    weights_df: pd.DataFrame = df[(df["xcat"] == weights) & cids_mask].copy()

    # set the targets and weights to wide indexing with cids as columns
    target_df = target_df.set_index(["real_date", "cid"])["value"].unstack(level=1)
    weights_df = weights_df.set_index(["real_date", "cid"])["value"].unstack(level=1)

    # Edge case where `weights` is the same as `xcat`, set weights to 1
    if weights is None or weights == "" or weights == xcat:
        weights_df = pd.DataFrame(
            data=np.ones(target_df.shape),
            index=target_df.index,
            columns=target_df.columns,
        )

    # Normalize the weights to sum to 1 if specified
    if normalize_weights:
        weights_df = weights_df.div(weights_df.abs().sum(axis=0), axis=1)

    # Downsample the weights to the update frequency
    weights_df = (
        weights_df.resample(update_freq)
        .mean(numeric_only=True)
        .reindex(target_df.index, method="bfill")
    )

    # Form a mask to apply NaNs where the weight or the target is NaN
    nan_mask = target_df.isna() | weights_df.isna()

    # Apply the weights to the target
    out_df = target_df * weights_df

    # Drop NaN cids as specified
    if complete_cids:
        out_df = out_df[out_df.columns[~nan_mask.any(axis=0)]]
    else:
        out_df = out_df[out_df.columns[~nan_mask.all(axis=0)]]

    out_df = out_df.sum(axis=1).reset_index().rename(columns={0: "value"})

    out_df["cid"] = new_cid
    out_df["xcat"] = xcat
    out_df = out_df.reset_index()[["cid", "xcat", "real_date", "value"]].sort_values(
        by=["cid", "xcat", "real_date"]
    )

    return out_df


def linear_composite(
    df: pd.DataFrame,
    xcats: Union[str, List[str]],
    cids: Optional[List[str]] = None,
    weights: Optional[Union[List[float], np.ndarray, pd.Series, str]] = None,
    update_freq: str = "M",
    signs: Optional[Union[List[float], np.ndarray, pd.Series]] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    complete_xcats: bool = True,
    complete_cids: bool = False,
    new_xcat="NEW",
    new_cid="GLB",
):
    """
    Returns new category panel as linear combination of others as standard dataframe

    :param <pd.DataFrame> df:  standardized JPMaQS DataFrame with the necessary
        columns: 'cid', 'xcat', 'real_date' and 'value'.
    :param <Union[str, List[str]> xcats: If a single category
        is given, the linear combination is calculated for all cross-sections available
        for that category - aggregating over cross-sections specified in `cids`,
        resulting in a single output series labelled with the new cross section name
        specified by `new_cid`.
        If a list of categories is given, the linear combination is calculated for all
        cross-sections available for each category specified in `xcats`, resulting in a
        panel of series labelled with the new category name specified by `new_xcat` and
        the same cross-sections as the input data.
    :param <List[str]> cids: cross-sections for which the linear combination is to be
        calculated. Default is all cross-section available.
    :param <Union[List[float], np.ndarray, pd.Series, str]> weights: An array of weights
        of the same length as the number of categories in `xcats` to be used in the
        linear combination. If a single category is given in `xcats`, another category
        can be specified in `weights` to be used as weights.
    :param <str> update_freq: The sampling frequency of the output data. The output
        data will be downsampled to the specified frequency. Options are 'D', 'W', 'M',
        'Q', 'A'. Default is 'M' (monthly).
    :param <List[float]> signs: An array of [1, -1] of the same length as the number of
        categories in `xcats` to indicate whether the respective category should be
        added or subtracted from the linear combination. Not relevant when aggregating
        over cross-sections, i.e. when a single category is given in `xcats`.
    :param <str> start: earliest date in ISO format. Default is None and earliest date
        for which the respective category is available is used.
    :param <str> end: latest date in ISO format. Default is None and latest date for
        which the respective category is available is used.
    :param <bool> complete_xcats: If True (default) combinations are only calculated for
        observation dates on which all xcats are available. If False a combination of the
        available categories is used. Not relevant when aggregating over cross-sections,
        i.e. when a single category is given in `xcats`.
    :param <bool> complete_cids: If True (default) combinations are only calculated for
        observation dates on which all cids are available. If False a combination of the
        available cross-sections is used. Not relevant when aggregating over categories,
        i.e. when multiple categories are given in `xcats`.
    :param <str> new_xcat: Name of new composite xcat when aggregating over xcats for a
        given cid. Default is "NEW".
    :param <str> new_cid: Name of new composite cid when aggregating over cids for a given
        xcat. Default is "GLB".

    :return <pd.DataFrame>: standardized DataFrame with the relative values, featuring
        the categories: 'cid', 'xcat', 'real_date' and 'value'.
    """
    listtypes = (list, np.ndarray, pd.Series)

    # df check
    if not isinstance(df, pd.DataFrame):
        raise TypeError("`df` must be a pandas DataFrame")

    if not set(["cid", "xcat", "real_date", "value"]).issubset(set(df.columns)):
        raise ValueError(
            "`df` must be a standardized JPMaQS DataFrame with the necessary columns: "
            "'cid', 'xcat', 'real_date' and 'value'."
        )

    if df.empty:
        raise ValueError("`df` is empty")

    dfx: pd.DataFrame = df.copy()

    # udpate_freq check
    if not isinstance(update_freq, str):
        raise TypeError("`update_freq` must be a string")

    update_freq = update_freq.upper()
    if update_freq not in ["D", "W", "M", "Q", "A"]:
        raise ValueError("`update_freq` must be one of 'D', 'W', 'M', 'Q', 'A'")

    # dates check
    for varx, namex in zip([start, end], ["start", "end"]):
        if varx is not None:
            if not isinstance(varx, str):
                raise TypeError(f"`{namex}` must be a string")
            if not is_valid_iso_date(varx):
                raise ValueError(f"`{namex}` must be a valid ISO date")

    dfx["real_date"] = pd.to_datetime(dfx["real_date"])
    if start is None:
        start = dfx["real_date"].min()
    else:
        start = pd.to_datetime(start)

    if end is None:
        end = dfx["real_date"].max()
    else:
        end = pd.to_datetime(end)

    # crop the df
    dfx = dfx[(dfx["real_date"] >= start) & (dfx["real_date"] <= end)]

    # now make sure the specified cids are in the df
    if cids is not None:
        if (
            (not isinstance(cids, list))
            or (not len(cids))
            or (not all(isinstance(x, str) for x in cids))
        ):
            raise TypeError("`cids` must be a non-empty list of strings")
        if not all(x in dfx["cid"].unique() for x in cids):
            raise ValueError(f"not all cids in `cids` are available in DataFrame")
    else:
        cids: List[str] = sorted(list(dfx["cid"].unique()))

    # Branch off for the single category case
    if isinstance(xcats, str):
        if not xcats in dfx["xcat"].unique():
            raise ValueError(f"Category '{xcats}' not available in DataFrame")

        if isinstance(weights, str):
            if not weights in dfx["xcat"].unique():
                raise ValueError(f"Category '{weights}' not available in DataFrame")
        else:
            raise TypeError(
                "`weights` must be a string specifying a category to be used as a weight"
            )

        tdf = dfx[dfx["cid"].isin(cids)]
        for cid in tdf["cid"].unique():
            avail_xcats: Set[str] = set(tdf[tdf["cid"] == cid]["xcat"].unique())
            if not xcats in avail_xcats:
                warnings.warn(
                    f"Category '{xcats}' (target category) not available for cid '{cid}',"
                    f"dropping cid from `cids`"
                )
                cids.remove(cid)
            if not weights in avail_xcats:
                warnings.warn(
                    f"Category '{weights}' (used as weights) not available for cid '{cid}',"
                    f"dropping cid from `cids`"
                )
                cids.remove(cid)

        dfx = reduce_df(dfx, cids=cids, xcats=[xcats, weights], start=start, end=end)

        return linear_composite_on_xcat(
            df=dfx,
            xcat=xcats,
            cids=cids,
            weights=weights,
            normalize_weights=False,
            complete_cids=complete_cids,
            new_cid=new_cid,
            update_freq=update_freq,
        )

    elif isinstance(xcats, list):
        if not all(isinstance(x, str) for x in xcats):
            raise TypeError("`xcats` must be a list of strings")
        if not all(x in dfx["xcat"].unique() for x in xcats):
            raise ValueError(f"Not all xcats in `xcats` are available in DataFrame")

        for varx, namex in zip([weights, signs], ["weights", "signs"]):
            if not isinstance(varx, listtypes):
                raise TypeError(
                    f"`{namex}` must be a list of floats, an np.ndarray or a pd.Series"
                )

        if weights is None:
            weights: np.ndarray = np.ones(len(xcats))

        if signs is None:
            signs: np.ndarray = np.ones(len(xcats))

        if not len(weights) == len(xcats) == len(signs):
            raise ValueError("`xcats`, `weights` and `signs` must have the same length")

        return linear_composite_on_cid(
            df=dfx,
            xcats=xcats,
            cids=cids,
            weights=weights,
            signs=signs,
            normalize_weights=True,
            complete_xcats=complete_xcats,
            new_xcat=new_xcat,
        )


if __name__ == "__main__":

    cids = ["AUD", "CAD", "GBP"]
    xcats = ["XR", "CRY", "INFL"]

    df: pd.DataFrame = pd.concat(
        [
            make_test_df(
                cids=cids,
                xcats=xcats[:-1],
                start_date="2000-01-01",
                end_date="2000-02-01",
                prefer="linear",
            ),
            make_test_df(
                cids=cids,
                xcats=["INFL"],
                start_date="2000-01-01",
                end_date="2000-02-01",
                prefer="decreasing-linear",
            ),
        ]
    )

    # all infls are now decreasing-linear, while everything else is increasing-linear

    lc_cid = linear_composite(
        df=df,
        xcats="CRY",
        weights="INFL",
    )

    lc_xcat = linear_composite(
        df=df,
        cids=["AUD", "CAD"],
        xcats=["XR", "CRY", "INFL"],
        weights=[1, 2, 1],
        signs=[1, -1, 1],
    )




import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import List, Tuple

from macrosynergy.management.simulate_quantamental_data import make_qdf
from macrosynergy.management.shape_dfs import reduce_df


def missing_in_df(df: pd.DataFrame, xcats: List[str] = None, cids: List[str] = None):
    """
    Print missing cross-sections and categories

    :param <pd.DataFrame> df: standardized DataFrame with the following necessary
        columns: 'cid', 'xcats', 'real_date'.
    :param <List[str]> xcats: extended categories to be checked on. Default is all
        in the DataFrame.
    :param <List[str]> cids: cross sections to be checked on. Default is all in
        the DataFrame.

    """
    print("Missing xcats across df: ", set(xcats) - set(df['xcat'].unique()))

    cids = df["cid"].unique() if cids is None else cids
    xcats_used = sorted(list(set(xcats).intersection(set(df["xcat"].unique()))))

    for xcat in xcats_used:
        cids_xcat = df.loc[df["xcat"] == xcat, "cid"].unique()
        print(f"Missing cids for {xcat}: ", set(cids) - set(cids_xcat))


def check_startyears(df: pd.DataFrame):
    """
    DataFrame with starting years across all extended categories and cross-sections

    :param <pd.DataFrame> df: standardized DataFrame with the following necessary
        columns: 'cid', 'xcats', 'real_date'.

    """
    df: pd.DataFrame = df.copy()
    df = df.dropna(how='any')
    df_starts = df[['cid', 'xcat', 'real_date']].groupby(['cid', 'xcat']).min()
    df_starts['real_date'] = pd.DatetimeIndex(df_starts.loc[:, 'real_date']).year

    return df_starts.unstack().loc[:, 'real_date'].astype(int, errors='ignore')


def check_enddates(df: pd.DataFrame):
    """
    DataFrame with end dates across all extended categories and cross sections.

    :param <pd.DataFrame> df: standardized DataFrame with the following necessary
        columns: 'cid', 'xcats', 'real_date'.
    """
    df: pd.DataFrame = df.copy()
    df = df.dropna(how='any')
    df_ends = df[['cid', 'xcat', 'real_date']].groupby(['cid', 'xcat']).max()
    df_ends['real_date'] = df_ends['real_date'].dt.strftime('%Y-%m-%d')

    return df_ends.unstack().loc[:, 'real_date']

def business_day_dif(df: pd.DataFrame, maxdate: pd.Timestamp):
    """
    Number of business days between two respective business dates.

    :param <pd.DataFrame> df: DataFrame cross-sections rows and category columns. Each
        cell in the DataFrame will correspond to the start date of the respective series.
    :param <pd.Timestamp> maxdate: maximum release date found in the received DataFrame.
        In principle, all series should have values up until the respective business
        date. The difference will represent possible missing values.

    :return <pd.DataFrame>: DataFrame consisting of business day differences for all
        series.

    """
    year_df = (maxdate.year - df.apply(lambda x: x.dt.isocalendar().year)) * 52
    week_df = (maxdate.week - df.apply(lambda x: x.dt.isocalendar().week))
    # Account for difference over a year.
    week_df += year_df
    # Account for weekends.
    week_df *= 2
    df = (maxdate - df).apply(lambda x: x.dt.days)
    return df - week_df

def visual_paneldates(df: pd.DataFrame, size: Tuple[float] = None):
    """
    Visualize panel dates with color codes.

    :param <pd.DataFrame> df: DataFrame cross sections rows and category columns.
    :param <Tuple[float]> size: tuple of floats with width/length of displayed heatmap.

    """

    # DataFrame of official timestamps.
    if all(df.dtypes == object):

        df = df.apply(pd.to_datetime)
        # All series, in principle, should be populated to the last active release date
        # in the DataFrame.
        maxdate = df.max().max()
        df = business_day_dif(df=df, maxdate=maxdate)

        df = df.astype(float)
        # Ideally the data type should be int, but Pandas cannot represent NaN as int.
        # -- https://pandas.pydata.org/pandas-docs/stable/user_guide/gotchas.html#support-for-integer-na

        header = f"Missing days up to {maxdate.strftime('%Y-%m-%d')}"

    else:

        header = "Start years of quantamental indicators."

    if size is None:
        size = (max(df.shape[0] / 2, 18), max(1, df.shape[1]/ 2))

    sns.set(rc={'figure.figsize': size})
    sns.heatmap(df.T, cmap='YlOrBr', center=df.stack().mean(), annot=True, fmt='.0f',
                linewidth=1, cbar=False)
    plt.xlabel("")
    plt.ylabel("")
    plt.title(header, fontsize=18)
    plt.show()


def check_availability(df: pd.DataFrame, xcats: List[str] = None,
                       cids: List[str] = None, start: str = None,
                       start_size: Tuple[float] = None, end_size: Tuple[float] = None,
                       start_years: bool = True, missing_recent: bool = True):
    """
    Wrapper for visualizing start and end dates of a filtered DataFrame.

    :param <pd.DataFrame> df: standardized DataFrame with the following necessary
        columns: 'cid', 'xcats', 'real_date'.
    :param <List[str]> xcats: extended categories to be checked on.
        Default is all in the DataFrame.
    :param <List[str]> cids: cross sections to be checked on.
        Default is all in the DataFrame.
    :param <str> start: string representing earliest considered date. Default is None.
    :param <Tuple[float]> start_size: tuple of floats with width / length of
        the start years heatmap. Default is None (format adjusted to data).
    :param <Tuple[float]> end_size: tuple of floats with width/length of
        the end dates heatmap. Default is None (format adjusted to data).
    :param <bool> start_years: boolean indicating whether or not to display a chart 
        of starting years for each cross-section and indicator.
        Default is True (display start years).
    :param <bool> missing_recent: boolean indicating whether or not to display a chart 
        of missing date numbers for each cross-section and indicator.
        Default is True (display missing days).
    """
    if not isinstance(start_years, bool):
        raise TypeError(f"<bool> object expected and not {type(start_years)}.")
    if not isinstance(missing_recent, bool):
        raise TypeError(f"<bool> object expected and not {type(missing_recent)}.")

    dfx = reduce_df(df, xcats=xcats, cids=cids, start=start)
    if start_years:
        dfs = check_startyears(dfx)
        visual_paneldates(dfs, size=start_size)
    if missing_recent:
        dfe = check_enddates(dfx)
        visual_paneldates(dfe, size=end_size)


if __name__ == "__main__":

    cids = ['AUD', 'CAD', 'GBP']
    xcats = ['XR', 'CRY']

    cols_1 = ['earliest', 'latest', 'mean_add', 'sd_mult']
    df_cids = pd.DataFrame(index=cids, columns=cols_1)
    df_cids.loc['AUD', ] = ['2010-01-01', '2020-12-31', 0.5, 2]
    df_cids.loc['CAD', ] = ['2010-01-01', '2020-11-30', 0, 1]
    df_cids.loc['GBP', ] = ['2012-01-01', '2020-11-30', -0.2, 0.5]

    cols_2 = cols_1 + ['ar_coef', 'back_coef']
    df_xcats = pd.DataFrame(index=xcats, columns=cols_2)
    df_xcats.loc['XR', ] = ['2010-01-01', '2020-12-31', 0, 1, 0, 0.3]
    df_xcats.loc['CRY', ] = ['2011-01-01', '2020-10-30', 1, 2, 0.9, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    filt_na = (dfd['cid'] == 'CAD') & (dfd['real_date'] < '2011-01-01')
    dfd.loc[filt_na, 'value'] = np.nan

    xxcats = xcats + ['TREND']
    xxcids = cids + ['USD']
    
    check_availability(df=dfd, xcats=xcats, cids=cids, start_size=(10, 5), end_size=(10, 8))


from .basket import Basket
from .category_relations import CategoryRelations
from .converge_row import ConvergeRow
from .return_beta import return_beta, beta_display
from .historic_vol import historic_vol
from .linear_composite import linear_composite
from .make_blacklist import make_blacklist
from .make_relative_value import make_relative_value
from .make_zn_scores import make_zn_scores
from .panel_calculator import panel_calculator
from .view_correlations import correl_matrix
from .view_grades import heatmap_grades
from .view_ranges import view_ranges
from .view_timelines import view_timelines
from .view_metrics import view_metrics

__all__ = ['Basket', 'CategoryRelations', 'ConvergeRow', 'return_beta',
           'beta_display', 'historic_vol', 'linear_composite',
           'make_blacklist', 'make_relative_value', 'make_zn_scores', 'panel_calculator',
           'correl_matrix', 'heatmap_grades', 'view_ranges', 'view_timelines', 'view_metrics']


"""Old DataQuery Interface path"""
import warnings
from ..download.jpmaqs import JPMaQSDownload

class Interface(JPMaQSDownload):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "api.Interface has been moved to macrosynergy.download.jpmaqs - will be removed in v0.1.0",
            DeprecationWarning,
            stacklevel=2
        )

        super().__init__(*args, **kwargs)
        
warnings.warn(
    "api.Interface has been moved to macrosynergy.download.jpmaqs - will be removed in v0.1.0",
    DeprecationWarning,
    stacklevel=2
)

__all__ = ["Interface"]



"""Macrosynergy Quant Research Package"""

try:
    from .version import git_revision as __git_revision__
    from .version import version as __version__
except ModuleNotFoundError:
    from setup import get_version_info

    FULLVERSION, GIT_REVISION = get_version_info()
    __version__ = FULLVERSION
    __git_revision__ = GIT_REVISION

__name__ = ["__version__"]


# allows the package version information to be accessed from the package


"""
Interface for downloading data from the JPMorgan DataQuery API.
This module is not intended to be used directly, but rather through
macrosynergy.download.jpmaqs.py. However, for a use cases independent
of JPMaQS, this module can be used directly to download data from the
JPMorgan DataQuery API.
"""
import concurrent.futures
import time
import os
import logging
import itertools
import base64
import uuid
import io
import warnings
import requests
from datetime import datetime, timedelta
from typing import List, Optional, Dict, Union, Tuple
from timeit import default_timer as timer
from tqdm import tqdm

from macrosynergy import __version__ as ms_version_info
from macrosynergy.download.exceptions import (
    AuthenticationError,
    DownloadError,
    InvalidResponseError,
    HeartbeatError,
    NoContentError,
    KNOWN_EXCEPTIONS,
)
from macrosynergy.management.utils import (
    is_valid_iso_date,
    form_full_url,
    Config,
)

CERT_BASE_URL: str = "https://platform.jpmorgan.com/research/dataquery/api/v2"
OAUTH_BASE_URL: str = (
    "https://api-developer.jpmorgan.com/research/dataquery-authe/api/v2"
)
OAUTH_TOKEN_URL: str = "https://authe.jpmchase.com/as/token.oauth2"
OAUTH_DQ_RESOURCE_ID: str = "JPMC:URI:RS-06785-DataQueryExternalApi-PROD"
JPMAQS_GROUP_ID: str = "JPMAQS"
API_DELAY_PARAM: float = 0.3  # 300ms delay between requests
API_RETRY_COUNT: int = 5  # retry count for transient errors
HL_RETRY_COUNT: int = 5  # retry count for "high-level" requests
MAX_CONTINUOUS_FAILURES: int = 5  # max number of continuous errors before stopping
HEARTBEAT_ENDPOINT: str = "/services/heartbeat"
TIMESERIES_ENDPOINT: str = "/expressions/time-series"
CATALOGUE_ENDPOINT: str = "/group/instruments"
HEARTBEAT_TRACKING_ID: str = "heartbeat"
OAUTH_TRACKING_ID: str = "oauth"
TIMESERIES_TRACKING_ID: str = "timeseries"
CATALOGUE_TRACKING_ID: str = "catalogue"

logger: logging.Logger = logging.getLogger(__name__)
debug_stream_handler = logging.StreamHandler(io.StringIO())
debug_stream_handler.setLevel(logging.NOTSET)
debug_stream_handler.setFormatter(
    logging.Formatter(
        "%(asctime)s - %(levelname)s - %(module)s - %(funcName)s :: %(message)s"
    )
)
logger.addHandler(debug_stream_handler)

egress_logger: dict = {}


def validate_response(
    response: requests.Response,
    user_id: str,
) -> dict:
    """
    Validates a response from the API. Raises an exception if the response
    is invalid (e.g. if the response is not a 200 status code).

    :param <requests.Response> response: response object from requests.request().

    :return <dict>: response as a dictionary. If the response is not valid,
        this function will raise an exception.

    :raises <InvalidResponseError>: if the response is not valid.
    :raises <AuthenticationError>: if the response is a 401 status code.
    :raises <KeyboardInterrupt>: if the user interrupts the download.
    """

    error_str: str = (
        f"Response: {response}\n"
        f"User ID: {user_id}\n"
        f"Requested URL: {response.request.url}\n"
        f"Response status code: {response.status_code}\n"
        f"Response headers: {response.headers}\n"
        f"Response text: {response.text}\n"
        f"Timestamp (UTC): {datetime.utcnow().isoformat()}; \n"
    )
    # TODO : Use response.raise_for_status() as a better way to check for errors
    if not response.ok:
        logger.info("Response status is NOT OK : %s", response.status_code)
        if response.status_code == 401:
            raise AuthenticationError(error_str)

        if HEARTBEAT_ENDPOINT in response.request.url:
            raise HeartbeatError(error_str)

        raise InvalidResponseError(
            f"Request did not return a 200 status code.\n{error_str}"
        )

    try:
        response_dict = response.json()
        if response_dict is None:
            raise InvalidResponseError(f"Response is empty.\n{error_str}")
        return response_dict
    except Exception as exc:
        if isinstance(exc, KeyboardInterrupt):
            raise exc

        raise InvalidResponseError(error_str + f"Error parsing response as JSON: {exc}")


def request_wrapper(
    url: str,
    headers: Optional[Dict] = None,
    params: Optional[Dict] = None,
    method: str = "get",
    tracking_id: Optional[str] = None,
    proxy: Optional[Dict] = None,
    cert: Optional[Tuple[str, str]] = None,
    **kwargs,
) -> dict:
    """
    Wrapper for requests.request() that handles retries and logging.
    All parameters and kwargs are passed to requests.request().

    :param <str> url: URL to request.
    :param <dict> headers: headers to pass to requests.request().
    :param <dict> params: params to pass to requests.request().
    :param <str> method: HTTP method to use. Must be one of "get"
        or "post". Defaults to "get".
    :param <dict> kwargs: kwargs to pass to requests.request().
    :param <str> tracking_id: default None, unique tracking ID of request.
    :param <dict> proxy: default None, dictionary of proxy settings for request.
    :param <Tuple[str, str]> cert: default None, tuple of string for filename
        of certificate and key.

    :return <dict>: response as a dictionary.

    :raises <InvalidResponseError>: if the response is not valid.
    :raises <AuthenticationError>: if the response is a 401 status code.
    :raises <DownloadError>: if the request fails after retrying.
    :raises <KeyboardInterrupt>: if the user interrupts the download.
    :raises <ValueError>: if the method is not one of "get" or "post".
    :raises <Exception>: other exceptions may be raised by requests.request().
    """

    if method not in ["get", "post"]:
        raise ValueError(f"Invalid method: {method}")

    user_id: str = kwargs.pop("user_id", "unknown")

    # insert tracking info in headers
    if headers is None:
        headers: Dict = {}
    headers["User-Agent"]: str = f"MacrosynergyPackage/{ms_version_info}"

    uuid_str: str = str(uuid.uuid4())
    if (tracking_id is None) or (tracking_id == ""):
        tracking_id: str = uuid_str
    else:
        tracking_id: str = f"uuid::{uuid_str}::{tracking_id}"

    headers["X-Tracking-Id"]: str = tracking_id

    log_url: str = form_full_url(url, params)
    logger.debug(f"Requesting URL: {log_url} with tracking_id: {tracking_id}")
    raised_exceptions: List[Exception] = []
    error_statements: List[str] = []
    error_statement: str = ""
    retry_count: int = 0
    response: Optional[requests.Response] = None
    upload_size: int = 0
    download_size: int = 0
    time_taken: float = 0
    start_time: float = timer()
    while retry_count < API_RETRY_COUNT:
        try:
            prepared_request: requests.PreparedRequest = requests.Request(
                method, url, headers=headers, params=params, **kwargs
            ).prepare()

            upload_size = prepared_request.headers.get("Content-Length", 0)

            response: requests.Response = requests.Session().send(
                prepared_request,
                proxies=proxy,
                cert=cert,
            )

            # track the download size
            if isinstance(response, requests.Response):
                download_size = response.content.__sizeof__()

            time_taken: float = timer() - start_time

            egress_logger[tracking_id] = {
                "url": log_url,
                "upload_size": int(upload_size),
                "download_size": int(download_size),
                "time_taken": time_taken,
            }

            if isinstance(response, requests.Response):
                return validate_response(response=response, user_id=user_id)

        except Exception as exc:
            # if keyboard interrupt, raise as usual
            if isinstance(exc, KeyboardInterrupt):
                print("KeyboardInterrupt -- halting download")
                raise exc

            # authentication error, clearly not a transient error
            if isinstance(exc, AuthenticationError):
                raise exc

            error_statement = (
                f"Request to {log_url} failed with error {exc}. "
                f"User ID: {user_id}. "
                f"Retry count: {retry_count}. "
                f"Tracking ID: {tracking_id}"
            )
            raised_exceptions.append(exc)
            error_statements.append(error_statement)

            known_exceptions = KNOWN_EXCEPTIONS + [HeartbeatError]
            # NOTE : HeartBeat is a special case

            # NOTE: exceptions that need the code to break should be caught before this
            # all other exceptions are caught here and retried after a delay

            if any([isinstance(exc, e) for e in known_exceptions]):
                egress_logger[tracking_id] = {
                    "url": log_url,
                    "upload_size": upload_size,
                    "download_size": download_size,
                    "time_taken": time_taken,
                    "error": f"{exc}",
                }
                logger.warning(error_statement)
                retry_count += 1
                time.sleep(API_DELAY_PARAM)
            else:
                raise exc

        time_taken = timer() - start_time

        egress_logger[tracking_id] = {
            "url": log_url,
            "upload_size": upload_size,
            "download_size": download_size,
            "time_taken": time_taken,
        }

    if isinstance(raised_exceptions[-1], HeartbeatError):
        raise HeartbeatError(error_statement)

    errs_str = "\n\n".join(
        ("\t" + str(e) + " - \n\t\t" + est)
        for e, est in zip(raised_exceptions, error_statements)
    )

    e_str = f"Request to {log_url} failed with error {raised_exceptions[-1]}. \n"
    e_str += "-" * 20 + "\n"
    if isinstance(response, requests.Response):
        e_str += f" Status code: {response.status_code}."
    e_str += (
        f" No longer retrying. Tracking ID: {tracking_id}"
        f"Exceptions raised:\n{errs_str}"
    )

    raise DownloadError(e_str)


class OAuth(object):
    """
    Class for handling OAuth authentication for the DataQuery API.

    :param <str> client_id: client ID for the OAuth application.
    :param <str> client_secret: client secret for the OAuth application.
    :param <dict> proxy: proxy to use for requests. Defaults to None.
    :param <str> token_url: URL for getting OAuth tokens.
    :param <str> dq_resource_id: resource ID for the JPMaQS Application.

    :return <OAuth>: OAuth object.

    :raises <ValueError>: if any of the parameters are semantically incorrect.
    :raises <TypeError>: if any of the parameters are of the wrong type.
    :raises <Exception>: other exceptions may be raised by underlying functions.
    """

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        proxy: Optional[dict] = None,
        token_url: str = OAUTH_TOKEN_URL,
        dq_resource_id: str = OAUTH_DQ_RESOURCE_ID,
    ):
        logger.debug("Instantiate OAuth pathway to DataQuery")
        vars_types_zip: zip = zip(
            [client_id, client_secret, token_url, dq_resource_id],
            [
                "client_id",
                "client_secret",
                "token_url",
                "dq_resource_id",
            ],
        )

        for varx, namex in vars_types_zip:
            if not isinstance(varx, str):
                raise TypeError(f"{namex} must be a <str> and not {type(varx)}.")

        if not isinstance(proxy, dict) and proxy is not None:
            raise TypeError(f"proxy must be a <dict> and not {type(proxy)}.")

        self.token_url: str = token_url
        self.proxy: Optional[dict] = proxy

        self._stored_token: Optional[dict] = None
        self.token_data = {
            "grant_type": "client_credentials",
            "client_id": client_id,
            "client_secret": client_secret,
            "aud": dq_resource_id,
        }

    def _valid_token(self) -> bool:
        """
        Method to check if the stored token is valid.

        :return <bool>: True if the token is valid, False otherwise.
        """
        if self._stored_token is None:
            logger.debug("No token stored")
            return False

        created: datetime = self._stored_token["created_at"]  # utc time of creation
        expires: datetime = created + timedelta(
            seconds=self._stored_token["expires_in"]
        )

        utcnow = datetime.utcnow()
        is_active: bool = expires > utcnow

        logger.debug(
            "Active token: %s, created: %s, expires: %s, now: %s",
            is_active,
            created,
            expires,
            utcnow,
        )

        return is_active

    def _get_token(self) -> str:
        """Method to get a new OAuth token.

        :return <str>: OAuth token.
        """
        if not self._valid_token():
            logger.debug("Request new OAuth token")
            js = request_wrapper(
                url=self.token_url,
                data=self.token_data,
                method="post",
                proxy=self.proxy,
                tracking_id=OAUTH_TRACKING_ID,
                user_id=self._get_user_id(),
            )
            # on failure, exception will be raised by request_wrapper

            # NOTE : use UTC time for token expiry
            self._stored_token: dict = {
                "created_at": datetime.utcnow(),
                "access_token": js["access_token"],
                "expires_in": js["expires_in"],
            }

        return self._stored_token["access_token"]
    
    def _get_user_id(self) -> str:
        return "OAuth_ClientID - " + self.token_data["client_id"]

    def get_auth(self) -> Dict[str, Union[str, Optional[Tuple[str, str]]]]:
        headers: Dict = {"Authorization": "Bearer " + self._get_token()}
        return {
            "headers": headers,
            "cert": None,
            "user_id": self._get_user_id(),
        }


class CertAuth(object):
    """
    Class for handling certificate based authentication for the DataQuery API.

    :param <str> username: username for the DataQuery API.
    :param <str> password: password for the DataQuery API.
    :param <str> crt: path to the certificate file.
    :param <str> key: path to the key file.

    :return <CertAuth>: CertAuth object.

    :raises <AssertionError>: if any of the parameters are of the wrong type.
    :raises <FileNotFoundError>: if certificate or key file is missing from filesystem.
    :raises <Exception>: other exceptions may be raised by underlying functions.
    """

    def __init__(
        self,
        username: str,
        password: str,
        crt: str,
        key: str,
        proxy: Optional[dict] = None,
    ):
        for varx, namex in zip([username, password], ["username", "password"]):
            if not isinstance(varx, str):
                raise TypeError(f"{namex} must be a <str> and not {type(varx)}.")

        self.auth: str = base64.b64encode(
            bytes(f"{username:s}:{password:s}", "utf-8")
        ).decode("ascii")

        # Key and Certificate check
        for varx, namex in zip([crt, key], ["crt", "key"]):
            if not isinstance(varx, str):
                raise TypeError(f"{namex} must be a <str> and not {type(varx)}.")
            if not os.path.isfile(varx):
                raise FileNotFoundError(f"The file '{varx}' does not exist.")
        self.key: str = key
        self.crt: str = crt
        self.username: str = username
        self.password: str = password
        self.proxy: Optional[dict] = proxy

    def get_auth(self) -> Dict[str, Union[str, Optional[Tuple[str, str]]]]:
        headers = {"Authorization": f"Basic {self.auth:s}"}
        user_id = "CertAuth_Username - " + self.username
        return {
            "headers": headers,
            "cert": (self.crt, self.key),
            "user_id": user_id,
        }


def validate_download_args(
    expressions: List[str],
    start_date: str,
    end_date: str,
    show_progress: bool,
    endpoint: str,
    calender: str,
    frequency: str,
    conversion: str,
    nan_treatment: str,
    reference_data: str,
    retry_counter: int,
    delay_param: float,
):
    """
    Validate the arguments passed to the download_data method.

    :params -- see download_data method.

    :returns True if all arguments are valid.

    :raises <TypeError>: if any of the arguments are of the wrong type.
    :raises <ValueError>: if any of the arguments are semantically incorrect.
    """

    if expressions is None:
        raise ValueError("`expressions` must be a list of strings.")

    if not isinstance(expressions, list):
        raise TypeError("`expressions` must be a list of strings.")

    if not all(isinstance(expr, str) for expr in expressions):
        raise TypeError("`expressions` must be a list of strings.")

    for varx, namex in zip([start_date, end_date], ["start_date", "end_date"]):
        if (varx is None) or not isinstance(varx, str):
            raise TypeError(f"`{namex}` must be a string.")
        if not is_valid_iso_date(varx):
            raise ValueError(
                f"`{namex}` must be a string in the ISO-8601 format (YYYY-MM-DD)."
            )

    if not isinstance(show_progress, bool):
        raise TypeError("`show_progress` must be a boolean.")

    if not isinstance(retry_counter, int):
        raise TypeError("`retry_counter` must be an integer.")

    if not isinstance(delay_param, float):
        raise TypeError("`delay_param` must be a float >=0.2 (seconds).")
    elif delay_param < 0.2:
        raise ValueError("`delay_param` must be a float >=0.2 (seconds).")

    vars_types_zip: zip = zip(
        [
            endpoint,
            calender,
            frequency,
            conversion,
            nan_treatment,
            reference_data,
        ],
        [
            "endpoint",
            "calender",
            "frequency",
            "conversion",
            "nan_treatment",
            "reference_data",
        ],
    )
    for varx, namex in vars_types_zip:
        if not isinstance(varx, str):
            raise TypeError(f"`{namex}` must be a string.")

    return True


def get_unavailable_expressions(
    expected_exprs: List[str],
    dicts_list: List[Dict],
) -> List[str]:
    """
    Method to get the expressions that are not available in the response.
    Looks at the dict["attributes"][0]["expression"] field of each dict
    in the list.

    :param <List[str]> expected_exprs: list of expressions that were requested.
    :param <List[Dict]> dicts_list: list of dicts to search for the expressions.

    :return <List[str]>: list of expressions that were not found in the dicts.
    """
    found_exprs: List[str] = [
        curr_dict["attributes"][0]["expression"]
        for curr_dict in dicts_list
        if curr_dict["attributes"][0]["time-series"] is not None
    ]
    return list(set(expected_exprs) - set(found_exprs))


class DataQueryInterface(object):
    """
    High level interface for the DataQuery API.
    Must be instantiated with a valid Config object.
    (see macrosynergy.management.utils.Config class for more info)

    :param <Config> config: Config object.
    :param <bool> oauth: whether to use OAuth authentication. Defaults to True.
    :param <bool> debug: whether to print debug messages. Defaults to False.
    :param <bool> concurrent: whether to use concurrent requests. Defaults to True.
    :param <int> batch_size: default 20, number of expressions to send in a single
        request. Must be a number between 1 and 20 (both included).
    :param <bool> check_connection: whether to send a check_connection request.
        Defaults to True.
    :param <str> base_url: base URL for the DataQuery API. Defaults to OAUTH_BASE_URL
        if `oauth` is True, CERT_BASE_URL otherwise.
    :param <str> token_url: token URL for the DataQuery API. Defaults to OAUTH_TOKEN_URL.
    :param <bool> suppress_warnings: whether to suppress warnings. Defaults to True.

    :return <DataQueryInterface>: DataQueryInterface object.

    :raises <TypeError>: if any of the parameters are of the wrong type.
    :raises <ValueError>: if any of the parameters are semantically incorrect.
    :raises <InvalidResponseError>: if the response from the server is not valid.
    :raises <DownloadError>: if the download fails to complete after a number of retries.
    :raises <HeartbeatError>: if the heartbeat (check connection) fails.
    :raises <Exception>: other exceptions may be raised by underlying functions.
    """

    def __init__(
        self,
        config: Config,
        oauth: bool = True,
        debug: bool = False,
        batch_size: int = 20,
        check_connection: bool = True,
        base_url: str = OAUTH_BASE_URL,
        token_url: str = OAUTH_TOKEN_URL,
        suppress_warnings: bool = True,
    ):
        self._check_connection: bool = check_connection
        self.msg_errors: List[str] = []
        self.msg_warnings: List[str] = []
        self.unavailable_expressions: List[str] = []
        self.debug: bool = debug
        self.suppress_warnings: bool = suppress_warnings
        self.batch_size: int = batch_size

        if not isinstance(config, Config):
            raise ValueError(
                "config_object must be provided for DataQuery authentication."
                " Check macrosynergy.management.utils.Config "
                "for more details."
            )
        self.config: Config = config
        self.auth: Optional[Union[CertAuth, OAuth]] = None
        if oauth and (config.oauth() is None):
            warnings.warn(
                "OAuth credentials not found. "
                "Trying to use certificate authentication.",)
            if config.cert() is None:
                raise ValueError(
                    "Certificate credentials not found. "
                    "Check the config_object passed."
                )
            else:
                oauth: bool = False
             
        if oauth:
            self.auth: OAuth = OAuth(
                **config.oauth(mask=False),
                token_url=token_url,
                proxy=config.proxy(mask=False),
            )
        else:
            if base_url == OAUTH_BASE_URL:
                base_url: str = CERT_BASE_URL

            self.auth: CertAuth = CertAuth(**config.cert(mask=False))

        assert (
            self.auth is not None
        ), "Failed to initialise access method. Check the config_object passed"

        self.proxy: Optional[dict] = config.proxy(mask=False)
        self.base_url: str = base_url
        self.egress_data: dict = {}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            logger.error("Exception %s - %s", exc_type, exc_value)
            print(f"Exception: {exc_type} {exc_value}")

    def check_connection(self, verbose=False) -> bool:
        """
        Check the connection to the DataQuery API using the Heartbeat endpoint.

        :param <bool> verbose: whether to print a message if the heartbeat
            is successful. Useful for debugging. Defaults to False.

        :return <bool>: True if the connection is successful, False otherwise.

        :raises <HeartbeatError>: if the heartbeat fails.
        """
        logger.debug("Check if connection can be established to JPMorgan DataQuery")
        js: dict = request_wrapper(
            url=self.base_url + HEARTBEAT_ENDPOINT,
            params={"data": "NO_REFERENCE_DATA"},
            proxy=self.proxy,
            tracking_id=HEARTBEAT_TRACKING_ID,
            **self.auth.get_auth(),
        )

        result: bool = True
        if (js is None) or (not isinstance(js, dict)) or ("info" not in js):
            logger.warning("Connection to JPMorgan DataQuery heartbeat failed")
            result: bool = False

        if result:
            result = (int(js["info"]["code"]) == 200) and (
                js["info"]["message"] == "Service Available."
            )

        if verbose:
            print("Connection successful!" if result else "Connection failed.")
        return result

    def _fetch(
        self,
        url: str,
        params: dict = None,
        tracking_id: Optional[str] = None,
    ) -> List[Dict]:
        """
        Make a request to the DataQuery API using the specified parameters.
        Used to wrap a request in a thread for concurrent requests, or to
        simplify the code for single requests.

        :param <str> url: URL to request.
        :param <dict> params: parameters to send with the request.
        :param <dict> proxy: proxy to use for the request.
        :param <str> tracking_id: tracking ID to use for the request.

        :return <List[Dict]>: list of dictionaries containing the response data.

        :raises <InvalidResponseError>: if the response from the server is not valid.
        :raises <Exception>: other exceptions may be raised by underlying functions.
        """

        downloaded_data: List[Dict] = []
        response: Dict = request_wrapper(
            url=url,
            params=params,
            proxy=self.proxy,
            tracking_id=tracking_id,
            **self.auth.get_auth(),
        )

        if (response is None) or ("instruments" not in response.keys()):
            if response is not None:
                if (
                    ("info" in response)
                    and ("code" in response["info"])
                    and (int(response["info"]["code"]) == 204)
                ):
                    raise NoContentError(
                        f"Content was not found for the request: {response}\n"
                        f"User ID: {self.auth.get_auth()['user_id']}\n"
                        f"URL: {form_full_url(url, params)}\n"
                        f"Timestamp (UTC): {datetime.utcnow().isoformat()}"
                    )

            raise InvalidResponseError(
                f"Invalid response from DataQuery: {response}\n"
                f"User ID: {self.auth.get_auth()['user_id']}\n"
                f"URL: {form_full_url(url, params)}"
                f"Timestamp (UTC): {datetime.utcnow().isoformat()}"
            )

        downloaded_data.extend(response["instruments"])

        if "links" in response.keys() and response["links"][1]["next"] is not None:
            logger.info("DQ response paginated - get next response page")
            downloaded_data.extend(
                self._fetch(
                    url=self.base_url + response["links"][1]["next"],
                    params={},
                    tracking_id=tracking_id,
                )
            )

        return downloaded_data

    def get_catalogue(
        self,
        group_id: str = JPMAQS_GROUP_ID,
    ) -> List[str]:
        """
        Method to get the JPMaQS catalogue.
        Queries the DataQuery API's Groups/Search endpoint to get the list of
        tickers in the JPMaQS group. The group ID can be changed to fetch a
        different group's catalogue.

        Parameters
        :param <str> group_id: the group ID to fetch the catalogue for.

        :return <List[str]>: list of tickers in the JPMaQS group.

        :raises <ValueError>: if the response from the server is not valid.
        """
        old_group_id: str = "CA_QI_MACRO_SYNERGY"
        try:
            response_list: Dict = self._fetch(
                url=self.base_url + CATALOGUE_ENDPOINT,
                params={"group-id": group_id},
                tracking_id=CATALOGUE_TRACKING_ID,
            )
        except NoContentError as e:
            response_list: Dict = self._fetch(
                url=self.base_url + CATALOGUE_ENDPOINT,
                params={"group-id": old_group_id},
                tracking_id=CATALOGUE_TRACKING_ID,
            )
        except Exception as e:
            raise e

        tickers: List[str] = [d["instrument-name"] for d in response_list]
        utkr_count: int = len(tickers)
        tkr_idx: List[int] = sorted([d["item"] for d in response_list])

        if not (
            (min(tkr_idx) == 1)
            and (max(tkr_idx) == utkr_count)
            and (len(set(tkr_idx)) == utkr_count)
        ):
            raise ValueError("The downloaded catalogue is corrupt.")

        return tickers

    def _download(
        self,
        expressions: List[str],
        params: dict,
        url: str,
        tracking_id: str,
        delay_param: float,
        show_progress: bool = False,
        retry_counter: int = 0,
    ) -> List[dict]:
        if retry_counter > 0:
            print("Retrying failed downloads. Retry count:", retry_counter)

        if retry_counter > HL_RETRY_COUNT:
            raise DownloadError(
                f"Failed {retry_counter} times to download data all requested data.\n"
                f"No longer retrying."
            )

        expr_batches: List[List[str]] = [
            expressions[i : i + self.batch_size]
            for i in range(0, len(expressions), self.batch_size)
        ]

        download_outputs: List[List[Dict]] = []
        failed_batches: List[List[str]] = []
        continuous_failures: int = 0
        last_five_exc: List[Exception] = []

        future_objects: List[concurrent.futures.Future] = []
        with concurrent.futures.ThreadPoolExecutor() as executor:
            for ib, expr_batch in tqdm(
                enumerate(expr_batches),
                desc="Requesting data",
                disable=not show_progress,
                total=len(expr_batches),
            ):
                curr_params: Dict = params.copy()
                curr_params["expressions"] = expr_batch
                future_objects.append(
                    executor.submit(
                        self._fetch,
                        url=url,
                        params=curr_params,
                        tracking_id=tracking_id,
                    )
                )
                time.sleep(delay_param)

            for ib, future in tqdm(
                enumerate(future_objects),
                desc="Downloading data",
                disable=not show_progress,
                total=len(future_objects),
            ):
                try:
                    download_outputs.append(future.result())
                    continuous_failures = 0
                except Exception as exc:
                    if isinstance(exc, (KeyboardInterrupt, AuthenticationError)):
                        raise exc

                    failed_batches.append(expr_batches[ib])
                    self.msg_errors.append(f"Batch {ib} failed with exception: {exc}")
                    continuous_failures += 1
                    last_five_exc.append(exc)
                    if continuous_failures > MAX_CONTINUOUS_FAILURES:
                        exc_str: str = "\n".join([str(e) for e in last_five_exc])
                        raise DownloadError(
                            f"Failed {continuous_failures} times to download data."
                            f" Last five exceptions: \n{exc_str}"
                        )

                    if self.debug:
                        raise exc

        final_output: List[Dict] = list(itertools.chain.from_iterable(download_outputs))

        if len(failed_batches) > 0:
            flat_failed_batches: List[str] = list(
                itertools.chain.from_iterable(failed_batches)
            )
            logger.warning(
                "Failed batches %d - retry download for %d expressions",
                len(failed_batches),
                len(flat_failed_batches),
            )
            retried_output: List[dict] = self._download(
                expressions=flat_failed_batches,
                params=params,
                url=url,
                tracking_id=tracking_id,
                delay_param=delay_param + 0.1,
                show_progress=show_progress,
                retry_counter=retry_counter + 1,
            )

            final_output += retried_output  # extend retried output

        return final_output

    def download_data(
        self,
        expressions: List[str],
        start_date: str = "2000-01-01",
        end_date: str = None,
        show_progress: bool = False,
        endpoint: str = TIMESERIES_ENDPOINT,
        calender: str = "CAL_ALLDAYS",
        frequency: str = "FREQ_DAY",
        conversion: str = "CONV_LASTBUS_ABS",
        nan_treatment: str = "NA_NOTHING",
        reference_data: str = "NO_REFERENCE_DATA",
        retry_counter: int = 0,
        delay_param: float = API_DELAY_PARAM,  # TODO do we want the user to have access to this?
    ) -> List[Dict]:
        """
        Download data from the DataQuery API.

        :param <List[str]> expressions: list of expressions to download.
        :param <str> start_date: start date for the data in the ISO-8601 format
            (YYYY-MM-DD).
        :param <str> end_date: end date for the data in the ISO-8601 format
            (YYYY-MM-DD).
        :param <bool> show_progress: whether to show a progress bar for the download.
        :param <str> endpoint: endpoint to use for the download.
        :param <str> calender: calendar setting to use for the download.
        :param <str> frequency: frequency of data points to use for the download.
        :param <str> conversion: conversion setting to use for the download.
        :param <str> nan_treatment: NaN treatment setting to use for the download.
        :param <str> reference_data: reference data to pass to the API kwargs.
        :param <int> retry_counter: number of times the download has been retried.
        :param <float> delay_param: delay between requests to the API.

        :return <List[Dict]>: list of dictionaries containing the response data.

        :raises <ValueError>: if any arguments are invalid or semantically incorrect
            (see validate_download_args()).
        :raises <DownloadError>: if the download fails.
        :raises <ConnectionError(HeartbeatError)>: if the heartbeat fails.
        :raises <Exception>: other exceptions may be raised by underlying functions.
        """
        tracking_id: str = TIMESERIES_TRACKING_ID
        if end_date is None:
            end_date = datetime.today().strftime("%Y-%m-%d")
            # NOTE : if "future dates" are passed, they must be passed by parent functions
            # see jpmaqs.py

        # NOTE : args validated only on first call, not on retries
        # this is because the args can be modified by the retry mechanism
        # (eg. date format)

        validate_download_args(
            expressions=expressions,
            start_date=start_date,
            end_date=end_date,
            show_progress=show_progress,
            endpoint=endpoint,
            calender=calender,
            frequency=frequency,
            conversion=conversion,
            nan_treatment=nan_treatment,
            reference_data=reference_data,
            retry_counter=retry_counter,
            delay_param=delay_param,
        )

        if datetime.strptime(end_date, "%Y-%m-%d") < datetime.strptime(
            start_date, "%Y-%m-%d"
        ):
            logger.warning(
                "Start date (%s) is after end-date (%s): swap them!",
                start_date,
                end_date,
            )
            start_date, end_date = end_date, start_date

        # remove dashes from dates to match DQ format
        start_date: str = start_date.replace("-", "")
        end_date: str = end_date.replace("-", "")

        # check heartbeat before each "batch" of requests
        if self._check_connection:
            if not self.check_connection():
                raise ConnectionError(
                    HeartbeatError(
                        f"Heartbeat failed. Timestamp (UTC):"
                        f" {datetime.utcnow().isoformat()}\n"
                        f"User ID: {self.auth.get_auth()['user_id']}\n"
                    )
                )
            time.sleep(delay_param)

        logger.info(
            "Download %d expressions from DataQuery from %s to %s",
            len(expressions),
            datetime.strptime(start_date, "%Y%m%d").date(),
            datetime.strptime(end_date, "%Y%m%d").date(),
        )
        params_dict: Dict = {
            "format": "JSON",
            "start-date": start_date,
            "end-date": end_date,
            "calendar": calender,
            "frequency": frequency,
            "conversion": conversion,
            "nan_treatment": nan_treatment,
            "data": reference_data,
        }

        final_output: List[dict] = self._download(
            expressions=expressions,
            params=params_dict,
            url=self.base_url + endpoint,
            tracking_id=tracking_id,
            delay_param=delay_param,
            show_progress=show_progress,
        )

        self.unavailable_expressions = get_unavailable_expressions(
            expected_exprs=expressions, dicts_list=final_output
        )
        logger.info(
            "Downloaded expressions: %d, unavailable: %d",
            len(final_output),
            len(self.unavailable_expressions),
        )

        self.egress_data = egress_logger
        return final_output


if __name__ == "__main__":
    import os

    cf: Config = Config(
        client_id=os.getenv("DQ_CLIENT_ID"),
        client_secret=os.getenv("DQ_CLIENT_SECRET"),
    )

    expressions = [
        "DB(JPMAQS,USD_EQXR_VT10,value)",
        "DB(JPMAQS,AUD_EXALLOPENNESS_NSA_1YMA,value)",
    ]

    with DataQueryInterface(config=cf) as dq:
        assert dq.check_connection(verbose=True)

        data = dq.download_data(
            expressions=expressions,
            start_date="2020-01-25",
            end_date="2023-02-05",
            show_progress=True,
        )

    print(f"Succesfully downloaded data for {len(data)} expressions.")




import numpy as np
import pandas as pd
from typing import List, Union, Tuple
import matplotlib.pyplot as plt
import math
import datetime as dt


class VintageData:
    """
    Creates standardized dataframe of single-ticker vintages. This class creates
    standardized grade 1 and grade 2 vintage data.

    :param <str> ticker: ticker name
    :param <str> cutoff: last possible release date. The format must be '%Y-%m-%d'.
            All other dates are calculated from this one. Default is end 2020.
    :param <list> release_lags: list of integers in ascending order denoting lags of the
        first, second etc. release in (calendar) days. Default is first release after 15
        days and revision after 30 days. If days fall on weekend they will be delayed to
        Monday.
    :param <int> number_firsts: number of first-release vintages in the simulated data
        set. Default is 24.
    :param <int> shortest: number of observations in the first (shortest) vintage.
        Default is 36.
    :param <str> freq: letter denoting the frequency of the vintage data. Must be one of
        'M' (monthly, default), 'Q' (quarterly)  or 'W' (weekly).
    :param <float> start_value: expected first value of the random series. Default is 100.
    :param <float> trend_ar: annualized trend. Default is 5% linear drift per year.
        This is applied to the start value. If the start value is not positive the
        linear trend is added as number.
    :param <float> sd_ar: annualized standard deviation. Default is sqrt(12).
    :param <float> seasonal: adds seasonal pattern (applying linear factor from low to
        high through the year) with value denoting the average % seasonal factor through
        the year. Default is None. The seasonal pattern makes only sense for values that
        are strictly positive and are interpreted as indices.
    :param <int> added_dates: number of added first release dates, used for grade 2
        dataframe generation. Default is 12.

    """

    def __init__(self, ticker, cutoff='2020-12-31', release_lags=[15, 30],
                 number_firsts=24, shortest=36, freq='M', start_value=100, trend_ar=5,
                 sd_ar=math.sqrt(12), seasonal=None, added_dates=12):

        self.ticker = ticker

        self.cutoff = self.date_check(cutoff)
        self.release_lags = release_lags
        self.number_firsts = number_firsts
        self.shortest = shortest

        self.freq = freq
        self.freq_int = dict(zip(['Q', 'M', 'W'], [4, 12, 52]))
        self.af = self.freq_int[freq]

        self.start_value = start_value
        self.trend_ar = trend_ar
        self.sd_ar = sd_ar
        self.seasonal = seasonal

        self.number_firsts_gr2 = self.number_firsts + added_dates
        self.dates = set()

    @staticmethod
    def date_check(date_string):
        """
        Validates that the dates passed are valid timestamp expressions and will convert
        to the required form '%Y-%m-%d'. Will raise an assertion if not in the expected
        form.

        :param <str> date_string: valid date expression. For instance, "1st January,
            2000."
        """
        date_error = "Expected form of string: '%Y-%m-%d'."
        if date_string is not None:
            try:
                pd.Timestamp(date_string).strftime("%Y-%m-%d")
            except ValueError:
                raise AssertionError(date_error)
            else:
                date = pd.Timestamp(date_string).strftime("%Y-%m-%d")
                return pd.Timestamp(date)

    @staticmethod
    def week_day(rel_date, day):

        if day == 0:
            return rel_date - dt.timedelta(days=1)

        return rel_date + dt.timedelta(days=1)

    def seasonal_adj(self, obs_dates, seas_factors, values):
        """
        Method used to seasonally adjust the series. Economic data can vary according to
        the season.

        :param <List[pd.Timestamps]> obs_dates: observation dates for the series.
        :param <List[float]> seas_factors: seasonal factors.
        :param <List[float> values: existing values that have not been seasonally
            adjusted.

        :return <List[float] values: returns a list of values which have been adjusted
            seasonally
        """
        if self.freq == 'W':
            week_dates = obs_dates.isocalendar().week
            condition = np.where(week_dates > 52)[0]
            if condition.size > 0:
                index = next(iter(condition))
                week_dates[index] = 52

            week = (week_dates - 1).to_numpy().astype(dtype=np.uint8)
            seasonal = seas_factors[week]
            return values * (1 + (seasonal / 100))

        month = (12 / self.af)
        month_freq = (obs_dates.month // month)
        month_freq = month_freq.astype(dtype=np.uint8)

        return values * (1 + seas_factors[month_freq - 1] / 100)

    def make_grade1(self):

        ref_date = self.cutoff - dt.timedelta(self.release_lags[0])
        ref_date = ref_date.replace(day=1)

        eop_dates = pd.date_range(end=ref_date, periods=self.number_firsts, freq=self.freq)
        eop_dates = eop_dates.date
        vin_lengths = list(range(self.shortest, self.shortest + self.number_firsts))
        v_first = vin_lengths[0]

        df_gr1 = pd.DataFrame(columns=['release_date', 'observation_date', 'value'])
        obs_dates = pd.date_range(end=eop_dates[0], periods=v_first, freq=self.freq)

        list_ = np.linspace(0, (self.af - 1), self.af)
        linear_scale = list_ - np.mean(list_)

        seas_factors = (self.seasonal / np.std(linear_scale))
        seas_factors *= linear_scale

        for i, eop_date in enumerate(eop_dates):

            v = vin_lengths[i]
            if i > 0:
                length = vin_lengths[i - 1]
                date = pd.Timestamp(eop_date)
                obs_dates = obs_dates.insert(loc=length, item=date)

            data = np.linspace(0, (v - 1), v)
            if self.start_value > 0:
                data = ((data * (self.trend_ar / self.af)) / 100)
                data = (1 + data)
                trend = self.start_value * data
            else:
                data = (data * (self.trend_ar / self.af))
                trend = self.start_value + data

            for rl in self.release_lags:
                rel_date = eop_date + dt.timedelta(days=rl)
                day = rel_date.weekday() + 5
                if day >= 0:
                    rel_date = self.week_day(rel_date, day)

                self.dates.add(rel_date)
                values = trend + np.random.normal(0, self.sd_ar / math.sqrt(self.af), v)
                if self.seasonal is not None:
                    values = self.seasonal_adj(obs_dates, seas_factors, values)

                df_rel = pd.DataFrame({'release_date': rel_date,
                                       'observation_date': obs_dates.date,
                                       'value': values})
                df_gr1 = pd.concat([df_gr1, df_rel], ignore_index=True)

        df_gr1["grading"] = 1
        return self.add_ticker_parts(df_gr1)

    def make_graded(self, grading, upgrades=[]):

        """
        Simulates an explicitly graded dataframe with a column 'grading'.

        :param <list> grading: optional addition of grading column. List of grades used
            from lowest to highest.
            Default is None. Must be a subset of [3, 2.2, 2.1, 1].
        :param <list> upgrades: indices of release dates at which the series upgrade.
            Must have length of grading minus one. Default is None.
        """

        assert len(upgrades) == (len(grading) - 1)
        df_grd = self.make_grade1()
        df_grd['grading'] = str(grading[0])
        dates = sorted(self.dates)
        for i, up in enumerate(upgrades):

            filt = df_grd['release_date'] >= dates[up]
            df_grd.loc[filt, 'grading'] = str(grading[(i + 1)])

        return df_grd

    @staticmethod
    def map_weekday(date):

        day = date.weekday() - 5
        if day >= 0:
            date = VintageData.week_day(date, day)

        return date

    def make_grade2(self):
        """
        Method used to construct a dataframe that consists of each respective observation
        date and the corresponding release date(s) (the release dates are computed using
        the observation date and the time-period(s) specified in the field
        "release_lags").

        :return <pd.DataFrame>: Will return the DataFrame with the additional columns.
        """
        ref_date = self.cutoff - dt.timedelta(self.release_lags[0])
        ref_date = ref_date.replace(day=1)

        eop_dates = pd.date_range(end=ref_date, periods=self.number_firsts_gr2,
                                  freq=self.freq)
        eop_list = eop_dates.date
        eop_arr = np.array(eop_list)

        df_gr2 = pd.DataFrame(columns = ['release_date', 'observation_date'])

        n_ends = len(eop_list)
        n_release = len(self.release_lags)

        data = np.zeros((n_ends, n_release), dtype=object)

        for i in range(n_release):
            dates = eop_arr + dt.timedelta(days=self.release_lags[i])
            dates = list(map(self.map_weekday, dates))
            data[:, i] = np.array(dates)

        shape = data.shape
        data = data.reshape((shape[0] * shape[1]))
        eop_arr = np.sort(np.tile(eop_arr, n_release))

        df_gr2['release_date'] = data
        df_gr2['observation_date'] = eop_arr
        df_gr2['ticker'] = self.ticker
        print(df_gr2)

        return self.add_ticker_parts(df_gr2)

    def add_ticker_parts(self, df):
        """
        Method used to add the associated tickers.

        :param <pd.DataFrame> df: standardised dataframe.

        :return <pd.DataFrame>: Will return the DataFrame with the additional columns.
        """
        old_cols = list(df.columns)
        add_cols = ['cross_section', 'category_code', 'adjustment', 'transformation']
        ticker_parts = self.ticker.split('_', 3)

        for i, tick in enumerate(ticker_parts):
            df[add_cols[i]] = tick

        if 'transformation' not in df.columns:
            df['transformation'] = None

        return df.loc[:, add_cols + old_cols]


if __name__ == "__main__":

    vins_m = VintageData('USD_INDX_SA', cutoff="2019-06-30", release_lags=[3, 20, 25],
                         number_firsts=12, shortest=12, sd_ar=5, trend_ar=20,
                         seasonal=10, added_dates=6)

    dfm1 = vins_m.make_grade1()

    dfm1.groupby('release_date').agg(['mean', 'count'])
    dfm2 = vins_m.make_grade2()
    dfmg = vins_m.make_graded(grading=[3, 2.1, 1], upgrades=[12, 24])

    vins_q = VintageData('USD_INDX_SA', release_lags=[3, 20, 25], number_firsts=2,
                         shortest=8, freq='Q', seasonal = 10, added_dates = 4)
    dfq1 = vins_q.make_grade1()
    dfq1.groupby('release_date').agg(['mean', 'count'])
    dfq2 = vins_q.make_grade2()

    vins_w = VintageData('USD_INDX_SA', cutoff="2019-06-30", release_lags=[3, 20, 25],
                         number_firsts=3 * 52, shortest=26, freq='W', seasonal = 10,
                         added_dates = 52)
    dfw1 = vins_w.make_grade1()
    dfw1.groupby('release_date').agg(['mean', 'count'])
    dfw2 = vins_w.make_grade2()

    dfm1.info()


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import List, Union, Tuple

from macrosynergy.management.simulate_quantamental_data import make_qdf
from macrosynergy.management.check_availability import reduce_df


def heatmap_grades(df: pd.DataFrame, xcats:  List[str],  cids: List[str] = None,
                   start: str = '2000-01-01', end: str = None, grade: str = 'grading',
                   title: str = None, size: Tuple[float] = None):

    """Displays a heatmap of grading

    :param <pd.Dataframe> df: standardized DataFrame with the necessary columns:
        'cid', 'xcat', 'real_date' and 'grading'.
    :param List[str] xcats: extended categorys to be checked on.
    :param <List[str]> cids: cross sections to visualize. Default is all in  DataFrame.
    :param <str> start: earliest date in ISO format. Default is earliest available.
    :param <str> end: latest date in ISO format. Default is latest available.
    :param <str> grade: name of column that contains the grades. Default is 'grading'.
    :param <str> title: string of chart title; defaults depend on type of range plot.
    :param <Tuple[float]> size: Tuple of width and height of graph.
        Default is None, meaning it is set in accordance with df.

    """
    df["real_date"] = pd.to_datetime(df["real_date"], format="%Y-%m-%d")

    df_cols = list(df.columns)
    grade_error = "Column that contains the grade values must be present in the " \
                  f"DataFrame: {df_cols}."
    assert grade in df.columns, grade_error

    df, xcats, cids = reduce_df(df, xcats, cids, start, end, out_all=True)
    df = df[['xcat', 'cid', 'real_date', grade]]
    df[grade] = df[grade].astype(float).round(2)

    df_ags = df.groupby(['xcat', 'cid']).mean().reset_index().\
        pivot(index='xcat', columns='cid', values=grade)

    if size is None:
        size = (max(df_ags.shape[0] / 2, 15), max(1, df_ags.shape[1] / 2))
    if title is None:
        sdate = df['real_date'].min().strftime('%Y-%m-%d')
        title = f'Average grade of vintages since {sdate}'
    sns.set(rc={'figure.figsize': size})
    sns.heatmap(df_ags, cmap='YlOrBr', vmin=1, vmax=3, annot=True, fmt='.1f',
                linewidth=1, cbar=False)
    plt.xlabel('')
    plt.ylabel('')
    plt.title(title, fontsize=18)
    plt.show()


if __name__ == "__main__":

    cids = ['NZD', 'AUD', 'CAD', 'GBP']
    xcats = ['XR', 'CRY', 'GROWTH', 'INFL']
    df_cids = pd.DataFrame(index=cids, columns=['earliest', 'latest',
                                                'mean_add', 'sd_mult'])
    df_cids.loc['AUD',] = ['2000-01-01', '2020-12-31', 0.1, 1]
    df_cids.loc['CAD',] = ['2001-01-01', '2020-11-30', 0, 1]
    df_cids.loc['GBP',] = ['2002-01-01', '2020-11-30', 0, 2]
    df_cids.loc['NZD',] = ['2002-01-01', '2020-09-30', -0.1, 2]

    df_xcats = pd.DataFrame(index=xcats, columns=['earliest', 'latest', 'mean_add',
                                                  'sd_mult', 'ar_coef', 'back_coef'])
    df_xcats.loc['XR',] = ['2000-01-01', '2020-12-31', 0.1, 1, 0, 0.3]
    df_xcats.loc['CRY',] = ['2000-01-01', '2020-10-30', 1, 2, 0.95, 1]
    df_xcats.loc['GROWTH',] = ['2001-01-01', '2020-10-30', 1, 2, 0.9, 1]
    df_xcats.loc['INFL',] = ['2001-01-01', '2020-10-30', 1, 2, 0.8, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    dfd['grading'] = '3'
    filter_date = dfd['real_date'] >= pd.to_datetime('2010-01-01')
    filter_cid = dfd['cid'].isin(['NZD', 'AUD'])
    dfd.loc[filter_date & filter_cid, 'grading'] = '1'
    filter_date = dfd['real_date'] >= pd.to_datetime('2013-01-01')
    filter_xcat = dfd['xcat'].isin(['CRY', 'GROWTH'])
    dfd.loc[filter_date & filter_xcat, 'grading'] = '2.1'
    filter_xcat = dfd['xcat'] == 'XR'
    dfd.loc[filter_xcat, 'grading'] = 1

    heatmap_grades(dfd, xcats=['CRY', 'GROWTH', 'INFL'], cids=cids)

    dfd.info()



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import List, Tuple, Callable, Optional
from packaging import version

from macrosynergy.management.simulate_quantamental_data import make_qdf
from macrosynergy.management.check_availability import reduce_df


def view_ranges(df: pd.DataFrame, xcats: List[str], cids: Optional[List[str]] = None,
                start: str = '2000-01-01', end: Optional[str] = None,
                val: str = 'value', kind: str = 'bar', 
                sort_cids_by: Optional[str] = None, 
                title: Optional[str] = None, ylab: Optional[str] = None, 
                size: Tuple[float] = (16, 8),
                xcat_labels: Optional[List[str]] = None):

    """Plots averages and various ranges across sections for one or more categories.

    :param <pd.Dataframe> df: standardized DataFrame with the necessary columns:
        'cid', 'xcats', 'real_date' and at least one column with values of interest.
    :param <List[str]> xcats: extended categories to be checked on. Default is all
        in the DataFrame.
    :param <List[str]> cids: cross sections to plot. Default is all in DataFrame.
    :param <str> start: earliest date in ISO format. Default earliest date in df.
    :param <str> end: latest date in ISO format. Default is latest date in df.
    :param <str> val: name of column that contains the values. Default is 'value'.
    :param <str> kind: type of range plot. Default is 'bar'; other option is 'box'.
    :param <str> sort_cids_by: criterion for sorting cids on x-axis;
        Arguments can be 'mean' and 'std'. Default is None, i.e. original order. Ordering
        will be based on the first category if the category is defined over the complete
        panel. Otherwise, mean and standard deviation calculated, of the cross-sections,
        computed across all categories.
    :param <str> title: string of chart title; defaults depend on type of range plot.
    :param <str> ylab: y label. Default is no label.
    :param <Tuple[float]> size: Tuple of width and height of graph. Default is (16, 8). 
    :param <List[str]> xcat_labels: custom labels to be used for the ranges.

    """

    df["real_date"] = pd.to_datetime(df["real_date"], format="%Y-%m-%d")

    possible_xcats = set(df["xcat"])
    missing_xcats = set(xcats).difference(possible_xcats)
    error_xcats = "The categories passed in to view_ranges() must be present in the " \
                  f"DataFrame: missing {missing_xcats}."
    if not set(xcats).issubset(possible_xcats):
        raise ValueError(error_xcats)

    sort_cids_func: Callable = None
    if sort_cids_by is not None:
        if not isinstance(sort_cids_by, str):
            raise TypeError("`sort_cids_by` must be a string.")
        sort_error = "Sorting parameter must either be 'mean' or 'std'."
        if not sort_cids_by in ['mean', 'std']:
            raise ValueError(sort_error)
        if sort_cids_by == "mean":
            sort_cids_func = np.mean
        else:
            sort_cids_func = np.std

    error_message = "The number of custom labels must match the defined number of " \
                    "categories in pnl_cats."
    if xcat_labels is not None:
        if (len(xcat_labels) != len(xcats)):
            raise ValueError(error_message)

    # Unique cross-sections across the union of categories passed - not the intersection.
    # Order of categories will be preserved.
    df, xcats, cids = reduce_df(df, xcats, cids, start, end, out_all=True)

    s_date = df['real_date'].min().strftime('%Y-%m-%d')
    e_date = df['real_date'].max().strftime('%Y-%m-%d')

    sns.set(style="darkgrid")

    if title is None:
        if kind == "bar":
            title = f"Means and standard deviations from {s_date} to {e_date}."
        elif kind == "box":
            title = f"Interquartile ranges, extended ranges and outliers " \
                    f"from {s_date} to {e_date}."
    if ylab is None:
        ylab = ""

    filt_1 = df['xcat'] == xcats[0]
    first_xcat_cids = set(df[filt_1]['cid'])
    # First category is not defined over all cross-sections.
    order_condition = list(set(cids)) == list(first_xcat_cids)

    if order_condition and sort_cids_func is not None:
        # Sort exclusively on the first category.
        dfx = df[filt_1].groupby(['cid'])[val].apply(sort_cids_func)
        order = dfx.sort_values(ascending=False).index
    elif not order_condition and sort_cids_func is not None:
        # Sort across all categories on the available cross-sections.
        dfx = df.groupby(['cid'])[val].apply(sort_cids_func)
        order = dfx.sort_values(ascending=False).index
    else:
        order = None

    sns.set_theme(style="darkgrid")
    fig, ax = plt.subplots(figsize=size)

    if kind == 'bar':
        if version.parse(sns.__version__) >= version.parse('0.12.0'):        
            ax = sns.barplot(
                x='cid', y=val, hue='xcat', hue_order=xcats,
                palette='Paired', data=df, errorbar='sd', order=order
                )
        else:
            ax = sns.barplot(x='cid', y=val, hue='xcat', hue_order=xcats,
                            palette='Paired', data=df, ci='sd', order=order)
    elif kind == 'box':
        ax = sns.boxplot(x='cid', y=val, hue='xcat', hue_order=xcats,
                         palette='Paired', data=df,  order=order)
        ax.xaxis.grid(True)

    ax.set_title(title,  fontdict={'fontsize': 16})
    ax.set_xlabel("")
    ax.set_ylabel(ylab)
    ax.xaxis.grid(True)
    ax.axhline(0, ls='--', linewidth=1, color='black')
    handles, labels = ax.get_legend_handles_labels()

    if xcat_labels is not None:
        error_message = "The number of custom labels must match the defined number of " \
                        "categories in pnl_cats."
        if (len(xcat_labels) != len(xcats)):
            raise ValueError(error_message)
        labels = xcat_labels

    if (len(xcats) == 1) and (xcat_labels is None):
        ax.get_legend().remove()
    else:
        ax.legend(handles=handles[0:], labels=labels[0:])
    plt.show()


if __name__ == "__main__":

    cids = ['AUD', 'CAD', 'GBP', 'USD']
    xcats = ['XR', 'CRY']
    df_cids = pd.DataFrame(index=cids, columns=['earliest', 'latest', 'mean_add',
                                                'sd_mult'])
    df_cids.loc['AUD', ] = ['2010-01-01', '2020-12-31', 0.5, 0.2]
    df_cids.loc['CAD', ] = ['2011-01-01', '2020-11-30', 0, 1]
    df_cids.loc['GBP', ] = ['2012-01-01', '2020-11-30', 0, 2]
    df_cids.loc['USD', ] = ['2012-01-01', '2020-11-30', 1, 2]

    df_xcats = pd.DataFrame(index=xcats, columns=['earliest', 'latest', 'mean_add',
                                                  'sd_mult', 'ar_coef', 'back_coef'])
    df_xcats.loc['XR', ] = ['2010-01-01', '2020-12-31', 0, 1, 0, 0.3]
    df_xcats.loc['CRY', ] = ['2011-01-01', '2020-10-30', 1, 2, 0.9, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    view_ranges(dfd, xcats=['XR'], kind='box', start='2012-01-01',
                end='2018-01-01', sort_cids_by='std')

    filter_1 = (dfd['xcat'] == 'XR') & (dfd['cid'] == 'AUD')
    dfd = dfd[~filter_1]

    view_ranges(dfd, xcats=['XR', 'CRY'], cids=cids, kind='box',
                start='2012-01-01', end='2018-01-01', sort_cids_by=None,
                xcat_labels=['EQXR_NSA', 'CRY_NSA'])




import itertools
import pandas as pd
import numpy as np
import seaborn as sns
import scipy.cluster.hierarchy as sch
from matplotlib import pyplot as plt
from typing import List, Union, Tuple
from collections import defaultdict

from macrosynergy.management.check_availability import reduce_df
from macrosynergy.management.simulate_quantamental_data import make_qdf

def lag_series(df_w: pd.DataFrame, lags: dict, xcats: List[str]):
    """
    Method used to lag respective categories.

    :param <pd.DataFrame> df_w: multi-index DataFrame where the columns are the
        categories, and the two indices are the cross-sections and real-dates.
    :param <dict> lags: dictionary of lags applied to respective categories.
    :param <List[str]> xcats: extended categories to be correlated.
    """

    lag_type = "The lag data structure must be of type <dict>."
    assert isinstance(lags, dict), lag_type

    lag_xcats = f"The categories referenced in the lagged dictionary must be " \
                f"present in the defined DataFrame, {xcats}."
    assert set(lags.keys()).issubset(set(xcats)), lag_xcats

    # Modify the dictionary to adjust for single categories having multiple lags.
    # The respective lags will be held inside a list.
    lag_copy = {}
    xcat_tracker = defaultdict(list)
    for xcat, shift in lags.items():
        if isinstance(shift, int):
            lag_copy[xcat + f"_L{shift}"] = shift
            xcat_tracker[xcat].append(xcat + f"_L{shift}")
        else:
            xcat_temp = [xcat + f"_L{s}" for s in shift]
            # Merge the two dictionaries.
            lag_copy = {**lag_copy, **dict(zip(xcat_temp, shift))}
            xcat_tracker[xcat].extend(xcat_temp)

    df_w_copy = df_w.copy()
    # Handle for multi-index DataFrame. The interior index represents the
    # timestamps.
    for xcat, shift in lag_copy.items():

        category = xcat[:-3]
        clause = isinstance(lags[category], list)
        first_lag = category in df_w.columns

        if clause and not first_lag:
            # Duplicate the column if multiple lags on the same category and the
            # category's first lag has already been implemented. Always access
            # the series from the original DataFrame.
            df_w[xcat] = df_w_copy[category]
        else:
            # Otherwise, modify the name.
            df_w = df_w.rename(columns={category: xcat})
        # Shift the respective column (name will have been adjusted to reflect
        # lag).
        df_w[xcat] = df_w.groupby(level=0)[xcat].shift(shift)

    return df_w, xcat_tracker

def correl_matrix(df: pd.DataFrame, xcats: Union[str, List[str]] = None,
                  cids: List[str] = None, start: str = '2000-01-01',
                  end: str = None, val: str = 'value', freq: str = None,
                  cluster: bool = False, lags: dict = None,
                  title: str = None, size: Tuple[float] = (14, 8),
                  max_color: float = None):
    """
    Visualize correlation across categories or cross-sections of panels.

    :param <pd.Dataframe> df: standardized JPMaQS DataFrame with the necessary columns:
        'cid', 'xcats', 'real_date' and at least one column with values of interest.
    :param <List[str]> xcats: extended categories to be correlated. Default is all in the
        DataFrame. If xcats contains only one category the correlation coefficients
        across cross sections are displayed. If xcats contains more than one category,
        the correlation coefficients across categories are displayed. Additionally, the
        order of the xcats received will be mirrored in the correlation matrix.
    :param <List[str]> cids: cross sections to be correlated. Default is all in the
        DataFrame.
    :param <str> start: earliest date in ISO format. Default is None and earliest date
        in df is used.
    :param <str> end: latest date in ISO format. Default is None and latest date in df
        is used.
    :param <str> val: name of column that contains the values of interest. Default is
        'value'.
    :param <str> freq: frequency option. Per default the correlations are calculated
        based on the native frequency of the datetimes in 'real_date', which is business
        daily. Down-sampling options include weekly ('W'), monthly ('M'), or quarterly
        ('Q') mean.
    :param <bool> cluster: if True the series in the correlation matrix are reordered
        by hierarchical clustering. Default is False.
    :param <dict> lags: optional dictionary of lags applied to respective categories.
        The key will be the category and the value is the lag or lags. If a
        category has multiple lags applied, pass in a list of lag values. The lag factor
        will be appended to the category name in the correlation matrix.
        N.B.: Lags can include a 0 if the original should also be correlated.
    :param <str> title: chart heading. If none is given, a default title is used.
    :param <Tuple[float]> size: two-element tuple setting width/height of figure. Default
        is (14, 8).
    :param <float> max_color: maximum values of positive/negative correlation
        coefficients for color scale. Default is none. If a value is given it applies
        symmetrically to positive and negative values.

    N.B:. The function displays the heatmap of a correlation matrix across categories or
    cross-sections (depending on which parameter has received multiple elements).
    """

    df["real_date"] = pd.to_datetime(df["real_date"], format="%Y-%m-%d")
    col_names = ['cid', 'xcat', 'real_date', val]
    df = df[col_names]

    if freq is not None:
        freq_options = ['W', 'M', 'Q']
        error_message = f"Frequency parameter must be one of the following options:" \
                        f"{freq_options}."
        assert freq in freq_options, error_message

    xcats = xcats if isinstance(xcats, list) else [xcats]

    if max_color is not None:
        assert isinstance(max_color, float), "Parameter max_color must be type <float>."

    min_color = None if max_color is None else -max_color

    df, xcats, cids = reduce_df(df, xcats, cids, start, end, out_all=True)

    s_date = df['real_date'].min().strftime('%Y-%m-%d')
    e_date = df['real_date'].max().strftime('%Y-%m-%d')

    if len(xcats) == 1:
        df_w = df.pivot(index='real_date', columns='cid', values=val)

        if freq is not None:
            df_w = df_w.resample(freq).mean()

        if title is None:
            title = f'Cross-sectional correlation of {xcats[0]} from {s_date} to ' \
                    f'{e_date}'

    else:

        df_w = df.pivot(index=('cid', 'real_date'), columns='xcat',
                        values=val)

        # Down-sample according to the passed frequency.
        if freq is not None:
            df_w = df_w.groupby([pd.Grouper(level='cid'),
                                 pd.Grouper(level='real_date', freq=freq)]
                                ).mean()

        # Apply the lag mechanism, to the respective categories, after the down-sampling.
        if lags is not None:
            df_w, xcat_tracker = lag_series(df_w=df_w, lags=lags, xcats=xcats)

        # Order the correlation DataFrame to reflect the order of the categories
        # parameter. Will replace the official category name with the lag appended name.
        if lags is not None:
            order = [[x] if x not in xcat_tracker.keys()
                     else xcat_tracker[x] for x in xcats]
            order = list(itertools.chain(*order))
        else:
            order = xcats

        df_w = df_w[order]

        if title is None:
            title = f'Cross-category correlation from {s_date} to {e_date}'

    sns.set(style="ticks")

    corr = df_w.corr(method='pearson')

    if cluster:
        # Pairwise distances between observations in n-dimensional space. If y is a 1-D
        # condensed distance matrix, then y must be a nCk sized vector (k = 2, pairwise).
        d = sch.distance.pdist(corr)
        # Perform hierarchical / agglomerative clustering. The clustering method used is
        # Farthest Point Algorithm.
        L = sch.linkage(d, method='complete')
        # Returns an (n - 1) by four matrix. The first two columns represent the "nodes"
        # being merged. The third column is the euclidean distance between the "nodes"
        # being merged and the fourth column equates to the number of original
        # observations in the newly formed cluster.
        # The second parameter is the distance threshold, t, which will determine the
        # "number" of clusters. If the distance threshold is too small, none of the data
        # points will form a cluster, so n different clusters are returned.
        # If there are any clusters, the categories contained in the cluster will be
        # adjacent.
        ind = sch.fcluster(L, 0.5 * d.max(), 'distance')
        columns = [corr.columns.tolist()[i] for i in list((np.argsort(ind)))]
        corr = corr.loc[columns, columns]

    # Mask for the upper triangle.
    # Return a copy of an array with the elements below the k-th diagonal zeroed. The
    # mask is implemented because correlation coefficients are symmetric.
    mask = np.triu(np.ones_like(corr, dtype=bool))

    fig, ax = plt.subplots(figsize=size)
    sns.heatmap(corr, mask=mask, cmap='vlag_r', center=0,
                vmin=min_color, vmax=max_color,
                square=False, linewidths=.5,
                cbar_kws={"shrink": .5})

    ax.set(xlabel='', ylabel='')
    ax.set_title(title, fontsize=14)

    plt.show()


if __name__ == "__main__":

    # Un-clustered correlation matrices.

    cids = ["AUD", "CAD", "GBP", "USD", "NZD", "EUR"]
    cids_dmsc = ["CHF", "NOK", "SEK"]
    cids_dmec = ["DEM", "ESP", "FRF", "ITL", "NLG"]
    cids += cids_dmec
    cids += cids_dmsc
    xcats = ["XR", "CRY"]

    df_cids = pd.DataFrame(index=cids, columns=['earliest', 'latest',
                                                'mean_add', 'sd_mult'])

    df_cids.loc['AUD'] = ['2010-01-01', '2020-12-31', 0.5, 2]
    df_cids.loc['CAD'] = ['2011-01-01', '2020-11-30', 0, 1]
    df_cids.loc['GBP'] = ['2012-01-01', '2020-11-30', -0.2, 0.5]
    df_cids.loc['USD'] = ['2010-01-01', '2020-12-30', -0.2, 0.5]
    df_cids.loc['NZD'] = ['2002-01-01', '2020-09-30', -0.1, 2]
    df_cids.loc['EUR'] = ['2002-01-01', '2020-09-30', -0.2, 2]
    df_cids.loc['DEM'] = ['2003-01-01', '2020-09-30', -0.3, 2]
    df_cids.loc['ESP'] = ['2003-01-01', '2020-09-30', -0.1, 2]
    df_cids.loc['FRF'] = ['2003-01-01', '2020-09-30', -0.2, 2]
    df_cids.loc['ITL'] = ['2004-01-01', '2020-09-30', -0.2, 0.5]
    df_cids.loc['NLG'] = ['2003-01-01', '2020-12-30', -0.1, 0.5]
    df_cids.loc['CHF'] = ['2003-01-01', '2020-12-30', -0.3, 2.5]
    df_cids.loc['NOK'] = ['2010-01-01', '2020-12-30', -0.1, 0.5]
    df_cids.loc['SEK'] = ['2010-01-01', '2020-09-30', -0.1, 0.5]

    df_xcats = pd.DataFrame(index=xcats, columns=['earliest', 'latest', 'mean_add',
                                                  'sd_mult', 'ar_coef', 'back_coef'])
    df_xcats.loc['XR', ] = ['2010-01-01', '2020-12-31', 0.1, 1, 0, 0.3]
    df_xcats.loc['CRY', ] = ['2010-01-01', '2020-10-30', 1, 2, 0.95, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    start = '2012-01-01'
    end = '2020-09-30'

    # Clustered correlation matrices. Test hierarchical clustering.
    correl_matrix(df=dfd, xcats='XR', cids=cids, start=start, end=end,
                  val='value', freq=None, cluster=True,
                  title='Correlation Matrix', size=(14, 8), max_color=None)

